#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
ConfigurationAdmin implementation

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

TODO: Stabilize implementation of managed service factories
FIXME: Add tests for the configuration of managed service factories
"""

# Standard library
import json
import logging
import os
import threading
import uuid

# Pelix
from pelix.ipopo.decorators import (
    ComponentFactory,
    Provides,
    Property,
    Validate,
    Invalidate,
    Requires,
    Instantiate,
    BindField,
    UnbindField,
)
import pelix.constants
import pelix.ldapfilter as ldapfilter
import pelix.services as services
import pelix.threadpool

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

SERVICE_CONFIGADMIN_DIRECTORY = "pelix.services.configadmin.directory"
""" Configurations directory service specification """

SERVICE_CONFIGURATION_ADMIN_PRIVATE = "pelix.services.configadmin.private"
"""
Private version of ConfigAdmin, to handle loop-requirements ("that will do")
"""

# ------------------------------------------------------------------------------


class Configuration(object):
    """
    Configuration object
    """

    def __init__(
        self, pid, properties, config_admin, persistence, factory_pid=None
    ):
        """
        :param pid: The configuration PID
        :param properties: The initial properties of the configuration
        :param config_admin: The parent ConfigurationAdmin service
        :param persistence: The configuration persistence handler
        :param factory_pid: An optional factory PID
        """
        # Configuration PID
        self.__pid = pid

        # Factory PID
        self.__factory_pid = factory_pid

        # Properties
        self.__properties = None
        self.__lock = threading.RLock()

        # Associated services
        self.__config_admin = config_admin
        self.__persistence = persistence

        # Configuration state
        self.__updated = False
        self.__deleted = False
        self.__location = None

        # Update using given properties, if any
        self.__properties_update(properties)

    def __str__(self):
        """
        String representation
        """
        if self.__factory_pid:
            kind = "FactoryConfiguration({0}, ".format(self.__factory_pid)
        else:
            kind = "Configuration("

        return "{0}pid={1}, updated={2}, deleted={3})".format(
            kind, self.__pid, self.__updated, self.__deleted
        )

    def get_bundle_location(self):
        """
        Get the bundle location.
        Returns the bundle location to which this configuration is bound,
        or None if it is not yet bound to a bundle location.

        :return: The location associated to the configuration
        """
        return self.__location

    def set_bundle_location(self, location):
        """
        Bind this Configuration object to the specified bundle location.
        If the location parameter is None then the Configuration object
        will not be bound to a location.
        It will be set to the bundle's location before the first time a
        Managed Service/Managed Service Factory receives this Configuration
        object via the updated method and before any plugins are called.
        The bundle location will be set persistently.

        :param location: A bundle location
        """
        self.__location = location

    def get_factory_pid(self):
        """
        For a factory configuration returns the PID of the corresponding
        Managed Service Factory, else returns None.

        :return: The factory PID or None
        """
        return self.__factory_pid

    def get_pid(self):
        """
        Returns the PID of this configuration

        :return: The configuration PID
        """
        return self.__pid

    def get_properties(self):
        """
        Return the properties of this Configuration object.
        The Dictionary object returned is a private copy for the caller and may
        be changed without influencing the stored configuration.

        If called just after the configuration is created and before update has
        been called, this method returns None.

        :return: A private copy of the properties for the caller or null.
                 These properties must not contain the "service.bundleLocation"
                 property. The value of this property may be obtained from the
                 get_bundle_location() method.
        """
        with self.__lock:
            if self.__deleted:
                raise ValueError("{0} has been deleted".format(self.__pid))

            elif not self.__updated:
                # Fresh configuration
                return None

            # Filter a copy of the properties
            props = self.__properties.copy()

            try:
                del props[services.CONFIG_PROP_BUNDLE_LOCATION]
            except KeyError:
                # Ignore
                pass

            return props

    def is_valid(self):
        """
        Checks if this configuration has been updated at least once and has not
        been deleted.

        :return: True if the configuration has properties and has not been
                 deleted
        """
        return self.__updated and not self.__deleted

    def __properties_update(self, properties):
        """
        Internal update of configuration properties. Does not notifies the
        ConfigurationAdmin of this modification.

        :param properties: the new set of properties for this configuration
        :return: True if the properties have been updated, else False
        """
        if not properties:
            # Nothing to do
            return False

        with self.__lock:
            # Make a copy of the properties
            properties = properties.copy()

            # Override properties
            properties[services.CONFIG_PROP_PID] = self.__pid

            if self.__location:
                properties[
                    services.CONFIG_PROP_BUNDLE_LOCATION
                ] = self.__location

            if self.__factory_pid:
                properties[
                    services.CONFIG_PROP_FACTORY_PID
                ] = self.__factory_pid

            # See if new properties are different
            if properties == self.__properties:
                return False

            # Store the copy (before storing data)
            self.__properties = properties
            self.__updated = True

            # Store the data
            # it will cause FileInstall to update this configuration again, but
            # this will ignored because self.__properties has already been
            # saved
            self.__persistence.store(self.__pid, properties)
            return True

    def reload(self):
        """
        Reloads the configuration file using the persistence service

        :raise IOError: File not found/readable
        :raise ValueError: Invalid file content
        """
        self.update(self.__persistence.load(self.__pid))

    def update(self, properties=None):
        # pylint: disable=W0212
        """
        If called without properties, only notifies listeners

        Update the properties of this Configuration object.
        Stores the properties in persistent storage after adding or overwriting
        the following properties:

        * "service.pid" : is set to be the PID of this configuration.
        * "service.factoryPid" : if this is a factory configuration it is set
          to the factory PID else it is not set.

        These system properties are all of type String.

        If the corresponding Managed Service/Managed Service Factory is
        registered, its updated method must be called asynchronously.
        Else, this callback is delayed until aforementioned registration
        occurs.

        Also initiates an asynchronous call to all ConfigurationListeners with
        a ConfigurationEvent.CM_UPDATED event.

        :param properties: the new set of properties for this configuration
        :raise IOError: Error storing the configuration
        """
        with self.__lock:
            # Update properties
            if self.__properties_update(properties):
                # Update configurations, if something changed
                self.__config_admin._update(self)

    def delete(self, directory_updated=False):
        # pylint: disable=W0212
        """
        Delete this configuration

        :param directory_updated: If True, tell ConfigurationAdmin to not
                                  recall the directory of this deletion
                                  (internal use only)
        """
        with self.__lock:
            if self.__deleted:
                # Nothing to do
                return

            # Update status
            self.__deleted = True

            # Notify ConfigurationAdmin, notify services only if the
            # configuration had been updated before
            self.__config_admin._delete(self, self.__updated, directory_updated)

            # Remove the file
            self.__persistence.delete(self.__pid)

            # Clean up
            if self.__properties:
                self.__properties.clear()

            self.__persistence = None
            self.__pid = None

    def matches(self, ldap_filter):
        """
        Tests if this configuration matches the given filter.

        :param ldap_filter: A parsed LDAP filter object
        :return: True if the properties of this configuration matches the
                 filter
        """
        if not self.is_valid():
            # Do not test invalid configurations
            return False

        return ldap_filter.matches(self.__properties)


# ------------------------------------------------------------------------------


@ComponentFactory()
@Provides(SERVICE_CONFIGADMIN_DIRECTORY)
@Requires("_admin", SERVICE_CONFIGURATION_ADMIN_PRIVATE)
@Instantiate("pelix-services-configuration-directory")
class _ConfigurationDirectory(object):
    """
    A configuration directory
    """

    def __init__(self):
        """
        Sets up members
        """
        # ConfigurationAdmin
        self._admin = None

        # PID -> Configuration
        self.__configurations = {}

        # Factory PIDs -> set(Configuration)
        self.__factories = {}

        # Lock
        self.__lock = threading.Lock()

    def exists(self, pid):
        """
        Checks if the given PID exists in the directory

        :param pid: A configuration PID
        :return: True if the PID already exists
        """
        return pid in self.__configurations or pid in self.__factories

    def get_configuration(self, pid):
        """
        Retrieves the configuration with the given PID

        :param pid: PID of the configuration
        :return: The configuration with the given PID
        :raise KeyError: Unknown PID
        """
        return self.__configurations[pid]

    def get_factory_configurations(self, factory_pid):
        """
        Retrieves the configurations with the given factory PID

        :param factory_pid: A factory PID
        :return: The set of matching configuration
        """
        return set(self.__factories.get(factory_pid, tuple()))

    def list_configurations(self, ldap_filter=None):
        """
        Returns the list of stored configurations

        :param ldap_filter: Optional LDAP filter
        :return: The set of matching configurations
        :raise ValueError: Invalid LDAP filter
        """
        if not ldap_filter:
            return set(self.__configurations.values())

        # Using an LDAP filter
        ldap_filter = ldapfilter.get_ldap_filter(ldap_filter)
        return {
            config
            for config in self.__configurations.values()
            if config.matches(ldap_filter)
        }

    def add(self, pid, properties, loader, factory_pid=None):
        """
        Creates a new configuration bean

        :param pid: PID of the configuration
        :param properties: Initial properties (can be None)
        :param loader: Persistence service associated to the configuration
        :param factory_pid: Set if the configuration is a factory (not used)
        :return: The new configuration bean
        :raise KeyError: PID already used
        :raise ValueError: Invalid PID or loader
        """
        with self.__lock:
            if pid in self.__configurations:
                raise KeyError("Already known configuration: {0}".format(pid))
            elif not pid:
                raise ValueError("Configuration with an empty PID")
            elif pid in self.__factories:
                raise KeyError(
                    "PID already used as a factory PID: {0}".format(pid)
                )
            elif loader is None:
                raise ValueError(
                    "No persistence service associated to {0}".format(pid)
                )

            # Make the configuration bean
            configuration = Configuration(
                pid, properties, self._admin, loader, factory_pid
            )

            # Store the factory according to the PID
            self.__configurations[pid] = configuration

            # Store the factory PID too
            if factory_pid is not None:
                self.__factories.setdefault(factory_pid, set()).add(
                    configuration
                )

            return configuration

    def update(self, pid, properties):
        """
        Updates the properties of an existing configuration

        :param pid: PID of a configuration
        :param properties: New properties (replace old dictionary)
        :raise KeyError: Unknown configuration
        :raise IOError: Error writing down the configuration
        """
        with self.__lock:
            # Update properties directly
            self.__configurations[pid].update(properties)

    def delete(self, pid):
        """
        Deletes the configuration with the given PID

        :param pid: PID of a configuration
        :raise KeyError: Unknown configuration
        :raise IOError: Error deleting the configuration file
        """
        with self.__lock:
            # Remove from the configuration dictionary
            config = self.__configurations.pop(pid)

            try:
                # Remove from the factory PIDs set
                factory_pid = config.get_factory_pid()
                factory_confs = self.__factories[factory_pid]
                factory_confs.remove(config)
                if not factory_confs:
                    del self.__factories[factory_pid]
            except KeyError:
                # Wasn't a factory configuration
                pass

            # Delete the configuration object
            config.delete(True)


# ------------------------------------------------------------------------------


@ComponentFactory()
@Provides(services.SERVICE_CONFIGURATION_ADMIN, controller="_controller")
@Provides(SERVICE_CONFIGURATION_ADMIN_PRIVATE)
@Requires("_directory", SERVICE_CONFIGADMIN_DIRECTORY, optional=True)
@Requires(
    "_persistences",
    services.SERVICE_CONFIGADMIN_PERSISTENCE,
    aggregate=True,
    optional=True,
)
@Requires(
    "_managed",
    services.SERVICE_CONFIGADMIN_MANAGED,
    aggregate=True,
    optional=True,
)
@Requires(
    "_managed_factories",
    services.SERVICE_CONFIGADMIN_MANAGED_FACTORY,
    aggregate=True,
    optional=True,
)
@Instantiate("pelix-services-configuration-admin")
class ConfigurationAdmin(object):
    """
    ConfigurationAdmin basic implementation
    """

    def __init__(self):
        # Service controller
        self._controller = False

        # Persistence services
        self._persistences = []

        # Managed services
        self._managed = []

        # Managed services factories
        self._managed_factories = []

        # Configurations directory
        self._directory = None

        # Service reference -> Managed Service
        self._managed_refs = {}
        self._factories_refs = {}

        # Some safety
        self.__lock = threading.RLock()

        # Update thread pool
        self._pool = None

        # Validation flag
        self.__validated = False

    def __set_up(self):
        """
        Set up the configuration administration service.
        To be called only when the component has been validated and the service
        controller is set to True.
        """
        # Get existing PIDs
        pids = set()
        for persistence in self._persistences:
            pids.update(persistence.get_pids())

        # Notify services
        self.__notify_pids(pids)

    @Validate
    def _validate(self, _):
        """
        Component validated
        """
        with self.__lock:
            # Create the update thread pool
            self._pool = pelix.threadpool.ThreadPool(2, logname="ConfigAdmin")
            self._pool.start()

            # Validation flag
            self.__validated = True

            # If the controller is on, set up the main service
            if self._controller:
                self.__set_up()

    @Invalidate
    def _invalidate(self, _):
        """
        Component invalidated
        """
        with self.__lock:
            # Validation flag
            self.__validated = False

            # Stop the pool
            self._pool.stop()
            self._pool = None

    @BindField("_directory")
    def _bind_directory(self, _, svc, svc_ref):
        # pylint: disable=W0613
        """
        The configurations directory has been bound
        """
        # Provide the ConfigurationAdmin service, if a controller is there
        self._controller = bool(self._persistences)
        if self.__validated and self._controller:
            # Set up the service
            self.__set_up()

    @BindField("_persistences")
    def _bind_persistence(self, _, svc, svc_ref):
        # pylint: disable=W0613
        """
        A persistence came in
        """
        self._controller = self._directory is not None
        if self.__validated and self._controller:
            # Set up the service
            self.__set_up()

    @UnbindField("_persistences")
    @UnbindField("_directory")
    def _unbind_directory(self, _, svc, svc_ref):
        # pylint: disable=W0613
        """
        The configurations directory has gone
        """
        # Remove the ConfigurationAdmin service
        self._controller = False

    @BindField("_managed")
    def _bind_managed(self, _, svc, svc_ref):
        """
        A managed service has been bound
        """
        with self.__lock:
            # Store the service reference
            self._managed_refs[svc_ref] = svc

            if self.__validated and self._controller:
                # Update with the associated configuration, if active
                pid = svc_ref.get_property(pelix.constants.SERVICE_PID)
                try:
                    self.__notify_single(pid, svc)
                except KeyError as ex:
                    _logger.error("Error configuring a service: %s", ex)

    @UnbindField("_managed")
    def _unbind_managed(self, _, svc, svc_ref):
        # pylint: disable=W0613
        """
        A managed service has gone
        """
        with self.__lock:
            # Forget the reference
            del self._managed_refs[svc_ref]

    @BindField("_managed_factories")
    def _bind_managed_factory(self, _, svc, svc_ref):
        """
        A managed service factory has been bound
        """
        with self.__lock:
            # Store the reference
            self._factories_refs[svc_ref] = svc

            if self.__validated and self._controller:
                # Update with associated configurations
                factory_pid = svc_ref.get_property(pelix.constants.SERVICE_PID)
                self.__notify_factory(factory_pid, svc)

    @UnbindField("_managed_factories")
    def _unbind_managed_factories(self, _, svc, svc_ref):
        # pylint: disable=W0613
        """
        A managed service has gone
        """
        with self.__lock:
            # Forget the reference
            del self._factories_refs[svc_ref]

    def __notify_pids(self, pids):
        """
        Updates the managed services for the given PIDs.

        This method should be called inside a locked block.

        :param pids: List of PIDs of configurations to load & update
        """
        for pid in pids:
            try:
                # Load the configuration
                config = self.get_configuration(pid)
                if config.is_valid():
                    # Notify corresponding service
                    self._update(config)
            except (IOError, ValueError) as ex:
                _logger.error("Error loading configuration %s: %s", pid, ex)

    def __get_matching_factories(self, factory_pid):
        """
        Returns the list of managed service factories that matches the given
        factory PID

        :param factory_pid: A managed service factory PID
        :return: The list of matching factories
        """
        return [
            svc
            for svc_ref, svc in self._factories_refs.items()
            if svc_ref.get_property(pelix.constants.SERVICE_PID) == factory_pid
        ]

    def __get_matching_services(self, pid):
        """
        Returns the list of managed services that matches the given PID

        :param pid: A configuration PID
        :return: The list of services matching the PID
        """
        # Make the list of managed services
        return [
            svc
            for svc_ref, svc in self._managed_refs.items()
            if svc_ref.get_property(pelix.constants.SERVICE_PID) == pid
        ]

    def __notify_single(self, pid, svc):
        """
        Adds a call to the updated() method of the given managed service in the
        pool, if a valid configuration has been found for it.

        :param pid: Service Persistent ID
        :param svc: Managed service
        """
        configuration = self.get_configuration(pid)
        if configuration.is_valid():
            # Valid configuration found, update the service
            self._pool.enqueue(svc.updated, configuration.get_properties())

    def __notify_factory(self, factory_pid, svc):
        """
        Adds a call to the updated() method of the given managed service
        factory in the pool, if valid configurations have been found for it.

        :param factory_pid: Factory Persistent ID
        :param svc: Managed service factory
        """
        configurations = self._directory.get_factory_configurations(factory_pid)
        if configurations:
            for configuration in configurations:
                if configuration.is_valid():
                    # Valid configurations found, call update for each one
                    self._pool.enqueue(
                        svc.updated,
                        configuration.get_pid(),
                        configuration.get_properties(),
                    )

    @staticmethod
    def __notify_factories(factories, pid, properties):
        """
        Calls the updated(pid, properties) method of managed service factories.

        :param factories: A list of managed service factories
        :param pid: PID of the deleted configuration
        :param properties: New configuration properties
        """
        for svc in factories:
            try:
                # Only give the properties to the service
                svc.updated(pid, properties)
            except Exception as ex:
                _logger.exception("Error updating factory: %s", ex)

    @staticmethod
    def __notify_factories_delete(factories, pid):
        """
        Calls the deleted(pid) method of the given managed service factories.

        :param factories: A list of managed service factories
        :param pid: PID of the deleted configuration
        """
        for svc in factories:
            try:
                svc.deleted(pid)
            except Exception as ex:
                _logger.exception("Error notifying a factory: %s", ex)

    @staticmethod
    def __notify_services(managed_services, properties):
        """
        Calls the updated(properties) method of managed services.
        Logs errors if necessary.

        :param managed_services: Managed services to be notified
        :param properties: New configuration properties
        """
        for svc in managed_services:
            try:
                # Only give the properties to the service
                svc.updated(properties)
            except Exception as ex:
                _logger.exception("Error updating service: %s", ex)

    def _update(self, configuration):
        """
        A configuration has been updated.

        Returns once managed services have been notified.

        :param configuration: The updated configuration
        """
        with self.__lock:
            future = None

            # Get configuration data
            factory_pid = configuration.get_factory_pid()
            pid = configuration.get_pid()
            properties = configuration.get_properties()

            if factory_pid:
                # Get the associated factories
                factories = self.__get_matching_factories(factory_pid)
                if factories:
                    # Call them from the pool
                    future = self._pool.enqueue(
                        self.__notify_factories, factories, pid, properties
                    )
            else:
                # Called corresponding managed services
                managed = self.__get_matching_services(configuration.get_pid())
                if managed:
                    # Call them from the pool
                    future = self._pool.enqueue(
                        self.__notify_services, managed, properties
                    )

        if future is not None:
            # Wait for the end of the notification, outside the lock
            future.result()

    def _delete(self, configuration, notify_services, directory_updated):
        """
        A configuration is about to be deleted.

        Returns once managed services have been notified.

        :param configuration: The deleted configuration
        :param notify_services: If True, notify services of the deletion
        :param directory_updated: If True, do not update the directory
        """
        with self.__lock:
            future = None

            factory_pid = configuration.get_factory_pid()
            pid = configuration.get_pid()

            # Remove the configuration from the directory
            if not directory_updated:
                self._directory.delete(pid)

            if notify_services:
                if factory_pid:
                    # Get the associated factories
                    factories = self.__get_matching_factories(factory_pid)
                    if factories:
                        # Call them from the pool
                        future = self._pool.enqueue(
                            self.__notify_factories_delete, factories, pid
                        )
                else:
                    # Called corresponding managed services
                    managed = self.__get_matching_services(pid)
                    if managed:
                        # Call them from the pool
                        future = self._pool.enqueue(
                            self.__notify_services, managed, None
                        )

        if future is not None:
            # Wait for the end of the notification, outside the lock
            future.result()

    def create_factory_configuration(self, factory_pid):
        """
        Create a new factory Configuration object with a new PID.
        The properties of the new Configuration object are null until the
        first time that its update() method is called.

        :param factory_pid: PID of the factory
        :raise ValueError: Invalid PID
        """
        with self.__lock:
            if factory_pid is None:
                raise ValueError("No factory PID given")

            try:
                # Remove leading-trailing spaces
                factory_pid = factory_pid.strip()
                if not factory_pid:
                    raise ValueError("Empty factory PID")
            except AttributeError:
                # .strip() doesn't exist
                raise ValueError(
                    "Invalid type of PID: {0}".format(
                        type(factory_pid).__name__
                    )
                )

            # Generate a configuration PID
            pid = "{0}-{1}".format(factory_pid, str(uuid.uuid4()))

            # Create the new factory configuration
            return self._directory.add(
                pid, None, self._persistences[0], factory_pid
            )

    def get_configuration(self, pid):
        """
        Get an existing Configuration object from the persistent store, or
        create a new Configuration object.

        :param pid: PID of the factory
        :raise IOError: File not found/readable
        """
        with self.__lock:
            try:
                return self._directory.get_configuration(pid)
            except KeyError:
                # Unknown configuration, look for it
                # (outside the exception block)
                pass

            for persistence in self._persistences[:]:
                if persistence.exists(pid):
                    # Load first existing one
                    properties = persistence.load(pid)
                    break
            else:
                # New configuration, with the best ranked persistence
                properties = {}
                persistence = self._persistences[0]

            # Take care of stored factory PID
            factory_pid = properties.get(services.CONFIG_PROP_FACTORY_PID)

            return self._directory.add(
                pid, properties, persistence, factory_pid
            )

    def list_configurations(self, ldap_filter=None):
        """
        List the current Configuration objects which match the filter.

        Only Configuration objects with non-null properties are considered
        current.
        That is, Configuration.get_properties() is guaranteed not to return
        null for each of the returned Configuration objects.

        The syntax of the filter string is as defined in the Filter class.
        The filter can test any configuration properties including the
        following:

        * service.pid (str): the PID under which this is registered
        * service.factoryPid (str): the factory if applicable
        * service.bundleLocation(str): the bundle location

        The filter can also be null, meaning that all Configuration objects
        should be returned.
        """
        with self.__lock:
            return self._directory.list_configurations(ldap_filter)


# ------------------------------------------------------------------------------


@ComponentFactory(services.FACTORY_CONFIGADMIN_JSON)
@Provides(
    [
        services.SERVICE_CONFIGADMIN_PERSISTENCE,
        services.SERVICE_FILEINSTALL_LISTENERS,
    ]
)
@Requires("_directory", SERVICE_CONFIGADMIN_DIRECTORY)
@Property("_conf_folder", "configuration.folder")
@Property("_watched_folder", services.PROP_FILEINSTALL_FOLDER)
class JsonPersistence(object):
    """
    JSON configuration persistence
    """

    def __init__(self):
        """
        Sets up members
        """
        # Configurations directory
        self._directory = None

        # Configuration folder
        self._conf_folder = None
        self._watched_folder = None

    def _get_file(self, pid):
        """
        Returns the path to the configuration file for the given PID

        :param pid: A configuration PID
        :return: The name of the configuration file
        """
        return os.path.join(self._conf_folder, "{0}.config.js".format(pid))

    @staticmethod
    def _get_pid(filename):
        """
        Extract the PID from the given file name

        :param filename: A file name
        :return: The corresponding PID or None
        """
        # Get the base name
        name = os.path.basename(filename)

        # Remove the extension
        try:
            ext_start = name.index(".config.js")
            return name[:ext_start] or None
        except IndexError:
            return None

    @Validate
    def validate(self, _):
        """
        Component validated
        """
        if not self._conf_folder:
            # ./conf is the default configuration folder
            self._conf_folder = os.path.join(os.getcwd(), "conf")

        # Make the folders if necessary
        if not os.path.exists(self._conf_folder):
            os.makedirs(self._conf_folder)

        # Set the folder watcher property
        self._watched_folder = self._conf_folder

    @Invalidate
    def invalidate(self, _):
        """
        Component invalidated
        """
        # Clear the cache
        self._watched_folder = None
        self._conf_folder = None

    def __load_file(self, filename):
        """
        Loads the configuration file with the given name

        :param filename: A simple file name
        :return: A tuple (PID, properties) or None
        """
        pid = self._get_pid(filename)
        if not pid:
            # Not a configuration file
            return None

        try:
            # Load the properties
            properties = self.load(pid)
        except IOError as ex:
            # Can't read file
            _logger.error("Error reading %s: %s", filename, ex)
            return None
        except ValueError as ex:
            # Bad JSON file
            _logger.error("Error parsing %s: %s", filename, ex)
            return None

        return pid, properties

    def exists(self, pid):
        """
        Returns True if a configuration with the given PID exists

        :param pid: PID of a configuration
        :return: True if a readable configuration exists
        """
        return os.path.isfile(self._get_file(pid))

    def load(self, pid):
        """
        Loads the configuration file for the given PID

        :param pid: A configuration PID
        :return: The properties in the configuration file
        :raise IOError: File not found/readable
        :raise ValueError: Invalid file content
        """
        with open(self._get_file(pid), "r") as filep:
            data = filep.read()

        # Store the configuration
        return json.loads(data)

    def store(self, pid, properties):
        """
        Stores the configuration with the given PID. Overwrites existing
        configuration.

        :param pid: A configuration PID
        :param properties: Configuration values
        :raise IOError: File not writable
        """
        # Write to the file
        with open(self._get_file(pid), "w") as filep:
            # Write the JSON data
            filep.write(
                json.dumps(
                    properties, sort_keys=True, indent=4, separators=(",", ": ")
                )
            )
            # Be nice, add a line feed
            filep.write("\n")

    def delete(self, pid):
        """
        Removes the configuration for the given PID. Does nothing if the
        configuration does not exist.

        :param pid: A configuration PID
        :return: True if the file has been successfully removed, else False
        """
        try:
            os.remove(self._get_file(pid))
            return True
        except OSError:
            return False

    def get_pids(self):
        """
        Returns the list of PIDs this storage could read
        """
        pids = set()
        for filename in os.listdir(self._conf_folder):
            if os.path.isfile(os.path.join(self._conf_folder, filename)):
                pid = self._get_pid(filename)
                if pid:
                    pids.add(pid)

        return pids

    def folder_change(self, folder, added, updated, deleted):
        """
        The configuration folder has been modified

        :param folder: Modified folder
        :param added: List of added files
        :param updated: List of modified files
        :param deleted: List of deleted files
        """
        # Check that the folder is really ours
        if folder != self._conf_folder:
            return

        # Handle deleted configurations first
        for filename in deleted:
            pid = self._get_pid(filename)
            try:
                # Delete the configuration
                self._directory.delete(pid)
            except KeyError:
                # Ignore unknown configuration
                pass

        # Handle updated configurations
        for filenames in added, updated:
            for filename in filenames:
                pid_props = self.__load_file(filename)
                if pid_props is None:
                    # File not readable
                    continue

                pid, properties = pid_props
                try:
                    try:
                        # Update the configuration
                        self._directory.update(pid, properties)
                    except KeyError:
                        # Configuration does not exist yet, create it
                        self._directory.add(pid, properties, self)
                except (KeyError, ValueError, IOError) as ex:
                    # Log other errors
                    _logger.error("Error updating %s: %s", pid, ex)


@pelix.constants.BundleActivator
class Activator(object):
    """
    Instantiates the default JSON configuration provider
    """

    @staticmethod
    def start(context):
        """
        Bundle started
        """
        if context.get_property(
            services.FRAMEWORK_PROP_CONFIGADMIN_DISABLE_DEFAULT_PERSISTENCE
        ):
            # No need to run the default provider
            return

        # Small trick to add a late Instantiate decoration
        # to the component factory
        Instantiate("pelix-services-configuration-json-default")(
            JsonPersistence
        )

    @staticmethod
    def stop(_):
        """
        Bundle stopped
        """



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Pelix OSGi-like services packages

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Service Registry Hooks

SERVICE_EVENT_LISTENER_HOOK = "pelix.internal.hooks.EventListenerHook"
# ------------------------------------------------------------------------------

FACTORY_EVENT_ADMIN = "pelix-services-eventadmin-factory"
""" Name of the EventAdmin component factory """

FACTORY_EVENT_ADMIN_MQTT = "pelix-services-eventadmin-mqtt-factory"
""" Name of the component factory of the MQTT bridge for EventAdmin """

# ------------------------------------------------------------------------------

SERVICE_EVENT_ADMIN = "pelix.services.eventadmin"
""" Specification of the EventAdmin service """

SERVICE_EVENT_HANDLER = "pelix.services.eventadmin.handler"
""" Specification of an EventAdmin event handler """

PROP_EVENT_TOPICS = "event.topics"
""" **List** of the topics handled by an event handler """

PROP_EVENT_FILTER = "event.filter"
""" Filter on events properties for an event handler """

EVENT_PROP_FRAMEWORK_UID = "event.sender.framework.uid"
""" UID of the framework that emitted the event """

EVENT_PROP_TIMESTAMP = "event.timestamp"
""" Time stamp of the event, compute during the call of send() or post() """

EVENT_PROP_PROPAGATE = "event.propagate"
"""
If present in event properties, the event can be propagated through MQTT
"""

# ------------------------------------------------------------------------------

SERVICE_CONFIGURATION_ADMIN = "pelix.configadmin"
""" Specification of the ConfigurationAdmin service """

SERVICE_CONFIGADMIN_MANAGED = "pelix.configadmin.managed"
""" Specification of a service managed by ConfigurationAdmin """

SERVICE_CONFIGADMIN_MANAGED_FACTORY = "pelix.configadmin.managed.factory"
""" Specification of a factory managed by ConfigurationAdmin """

SERVICE_CONFIGADMIN_PERSISTENCE = "pelix.configadmin.persistence"
""" Specification of a ConfigurationAdmin storage service """

FRAMEWORK_PROP_CONFIGADMIN_DISABLE_DEFAULT_PERSISTENCE = (
    "pelix.configadmin.persistence.default.disable"
)
"""
If this framework property has a value, the default persistence service of
ConfigurationAdmin won't be started
"""

FACTORY_CONFIGADMIN_JSON = "pelix-configadmin-persistence-json-factory"
""" Name of the JSON ConfigurationAdmin storage component factory """

CONFIG_PROP_PID = "service.pid"
""" Configuration property: the configuration PID """

CONFIG_PROP_FACTORY_PID = "service.factoryPid"
""" Configuration property: factory PID (not used yet) """

CONFIG_PROP_BUNDLE_LOCATION = "service.bundleLocation"
""" Configuration property: bound location (not used yet) """

# ------------------------------------------------------------------------------

SERVICE_FILEINSTALL = "pelix.services.fileinstall"
""" Specification of the File Install service """

SERVICE_FILEINSTALL_LISTENERS = "pelix.services.fileinstall.listener"
""" Specification of a listener of the File Install service """

PROP_FILEINSTALL_FOLDER = "fileinstall.folder"
""" Path to the folder to look after, in white board pattern """

# ------------------------------------------------------------------------------

SERVICE_MQTT_CONNECTOR_FACTORY = "pelix.mqtt.factory"
""" Specification of an MQTT connection factory """

MQTT_CONNECTOR_FACTORY_PID = "mqtt.connector"
""" PID of the MQTT connection factory """

SERVICE_MQTT_CONNECTION = "pelix.mqtt.connection"
""" Specification of an MQTT connection service """

SERVICE_MQTT_LISTENER = "pelix.mqtt.listener"
""" Specification of an MQTT message listener """

PROP_MQTT_TOPICS = "pelix.mqtt.topics"
""" List of the topics a listener wants to subscribes to """



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
An EventAdmin handler which prints to the standard output the events it receives

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
from pprint import pformat
import logging

# Pelix
from pelix.ipopo.decorators import (
    ComponentFactory,
    Provides,
    Property,
    Validate,
)
import pelix.misc
import pelix.services as services

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


def _parse_boolean(value):
    """
    Returns a boolean value corresponding to the given value.

    :param value: Any value
    :return: Its boolean value
    """
    if not value:
        return False

    try:
        # Lower string to check known "false" value
        value = value.lower()
        return value not in ("none", "0", "false", "no")
    except AttributeError:
        # Not a string, but has a value
        return True


# ------------------------------------------------------------------------------


@ComponentFactory(pelix.misc.FACTORY_EVENT_ADMIN_PRINTER)
@Provides(services.SERVICE_EVENT_HANDLER)
@Property("_event_topics", services.PROP_EVENT_TOPICS, "*")
@Property("_print", "evt.print", True)
@Property("_log", "evt.log", False)
class EventAdminPrinter(object):
    # pylint: disable=R0903
    """
    Utility component which can print and log EventAdmin events
    """

    def __init__(self):
        """
        Sets up members
        """
        self._event_topics = None
        self._print = False
        self._log = False

    @Validate
    def _validate(self, _):
        """
        Component validated
        """
        # Normalize parameters
        self._print = _parse_boolean(self._print)
        self._log = _parse_boolean(self._log)

    def handle_event(self, topic, properties):
        """
        An EventAdmin event has been received
        """
        if self._print:
            # Print the event on standard output
            print(
                "Event: {0}\nProperties:\n{1}".format(
                    topic, pformat(properties)
                )
            )

        if self._log:
            # Log the event
            _logger.info("Event: %s ; Properties: %s", topic, properties)



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Implementation of the OSGi LogService, based on Python standard logging

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import collections
import datetime
import logging
import sys
import time
import traceback

# Pelix
import pelix.framework
from pelix.constants import BundleActivator
from pelix.misc import (
    LOG_SERVICE,
    LOG_READER_SERVICE,
    PROPERTY_LOG_LEVEL,
    PROPERTY_LOG_MAX_ENTRIES,
)

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# Local logger
logger = logging.getLogger(__name__)

# Definition of the log levels (OSGi values)
LOG_ERROR = 1
LOG_WARNING = 2
LOG_INFO = 3
LOG_DEBUG = 4

# OSGi level => Python logging level
OSGI_TO_LEVEL = {
    LOG_DEBUG: logging.DEBUG,
    LOG_INFO: logging.INFO,
    LOG_WARNING: logging.WARNING,
    LOG_ERROR: logging.ERROR,
}

# Python logging level => OSGi level
LEVEL_TO_OSGI = {
    logging.DEBUG: LOG_DEBUG,
    logging.INFO: LOG_INFO,
    logging.WARNING: LOG_WARNING,
    logging.ERROR: LOG_ERROR,
    logging.CRITICAL: LOG_ERROR,
}

# ------------------------------------------------------------------------------


class LogEntry(object):
    """
    Represents a log entry
    """

    __slots__ = (
        "__bundle",
        "__exception",
        "__level",
        "__message",
        "__reference",
        "__time",
        "__record",
    )

    def __init__(self, level, message, exception, bundle, reference):
        """
        :param level: The Python log level of the entry
        :param message: A human readable message
        :param exception: The exception associated to the entry
        :param bundle: The bundle that created the entry
        :param reference: The service reference associated to the entry
        """
        self.__bundle = bundle
        self.__exception = exception
        self.__level = level
        self.__message = message
        self.__reference = reference
        self.__time = time.time()
        self.__record = None

    def __str__(self):
        """
        String representation
        """
        values = [
            # 7: length of "WARNING"
            "{0: ^7} ::".format(logging.getLevelName(self.__level)),
            # Date
            str(datetime.datetime.fromtimestamp(self.__time)),
            "::",
        ]

        if self.__bundle:
            # Bundle name
            values.append(
                "{0: <20s} ::".format(self.__bundle.get_symbolic_name())
            )

        # Message
        values.append(self.__message)

        if not self.__exception:
            # Print as is
            return " ".join(values)

        # Print the exception too
        return "{0}\n{1}".format(" ".join(values), self.__exception)

    @property
    def bundle(self):
        """
        The bundle that created this entry
        """
        return self.__bundle

    @property
    def message(self):
        """
        The message associated to this entry
        """
        return self.__message

    @property
    def exception(self):
        """
        The exception associated to this entry
        """
        return self.__exception

    @property
    def level(self):
        """
        The log level of this entry (Python constant)
        """
        return self.__level

    @property
    def osgi_level(self):
        """
        The log level of this entry (OSGi constant)
        """
        return LEVEL_TO_OSGI.get(self.__level, LOG_INFO)

    @property
    def reference(self):
        """
        The reference to the service associated to this entry
        """
        return self.__reference

    @property
    def time(self):
        """
        The timestamp of this entry
        """
        return self.__time

    def to_record(self):
        # type: () -> logging.LogRecord
        """
        Returns this object as a ``logging.LogRecord``
        """
        if self.__record is None:
            # Construct the record on demand
            self.__record = self.__make_record()

        return self.__record

    def __make_record(self):
        """
        Converts this object into a ``logging.LogRecord`` object
        """
        # Extract local details
        bundle = self.bundle
        name = bundle.get_symbolic_name()
        pathname = bundle.get_location()
        lineno = 0

        args = []
        func = "n/a"
        sinfo = None

        level = self.level
        msg = self.message
        exc_info = self.exception

        # Construct the record
        record = logging.LogRecord(
            name, level, pathname, lineno, msg, args, exc_info, func, sinfo
        )

        # Fix the time related entries
        log_start_time = record.created - (record.relativeCreated / 1000)
        creation_time = self.__time

        record.created = creation_time
        record.msecs = (creation_time - int(creation_time)) * 1000
        record.relativeCreated = (creation_time - log_start_time) * 1000

        return record


class LogReaderService:
    """
    The LogReader service
    """

    def __init__(self, context, max_entries):
        """
        :param context: The bundle context
        :param max_entries: Maximum stored entries
        """
        self._context = context
        self.__logs = collections.deque(maxlen=max_entries)
        self.__listeners = set()

    def add_log_listener(self, listener):
        """
        Subscribes a listener to log events.

        A log listener is an object providing with a ``logged`` method, with
        the following signature:

        .. code-block:: python

            def logged(self, log_entry):
                '''
                A log entry (LogEntry) has been added to the log service
                '''
                # ...

        :param listener: A new listener
        """
        if listener is not None:
            self.__listeners.add(listener)

    def remove_log_listener(self, listener):
        """
        Unsubscribes a listener from log events.

        :param listener: The listener to remove
        """
        self.__listeners.discard(listener)

    def get_log(self):
        """
        Returns the logs events kept by the service

        :return: A tuple of log entries
        """
        return tuple(self.__logs)

    def _store_entry(self, entry):
        """
        Stores a new log entry and notifies listeners

        :param entry: A LogEntry object
        """
        # Get the logger and log the message
        self.__logs.append(entry)

        # Notify listeners
        for listener in self.__listeners.copy():
            try:
                listener.logged(entry)
            except Exception as ex:
                # Create a new log entry, without using logging nor notifying
                # listener (to avoid a recursion)
                err_entry = LogEntry(
                    logging.WARNING,
                    "Error notifying logging listener {0}: {1}".format(
                        listener, ex
                    ),
                    sys.exc_info(),
                    self._context.get_bundle(),
                    None,
                )

                # Insert the new entry before the real one
                self.__logs.pop()
                self.__logs.append(err_entry)
                self.__logs.append(entry)


class LogServiceInstance:
    # pylint: disable=R0903
    """
    Instance of the log service given to a bundle by the factory
    """

    __slots__ = ("__reader", "__bundle")

    def __init__(self, reader, bundle):
        """
        :param reader: The Log Reader service
        :param bundle: Bundle associated to this instance
        """
        self.__reader = reader
        self.__bundle = bundle

    def log(self, level, message, exc_info=None, reference=None):
        # pylint: disable=W0212
        """
        Logs a message, possibly with an exception

        :param level: Severity of the message (Python logging level)
        :param message: Human readable message
        :param exc_info: The exception context (sys.exc_info()), if any
        :param reference: The ServiceReference associated to the log
        """
        if not isinstance(reference, pelix.framework.ServiceReference):
            # Ensure we have a clean Service Reference
            reference = None

        if exc_info is not None:
            # Format the exception to avoid memory leaks
            try:
                exception_str = "\n".join(traceback.format_exception(*exc_info))
            except (TypeError, ValueError, AttributeError):
                exception_str = "<Invalid exc_info>"
        else:
            exception_str = None

        # Store the LogEntry
        entry = LogEntry(
            level, message, exception_str, self.__bundle, reference
        )
        self.__reader._store_entry(entry)


class LogServiceFactory(logging.Handler):
    """
    Log Service Factory: provides a logger per bundle
    """

    def __init__(self, context, reader, level):
        """
        :param context: The bundle context
        :param reader: The Log Reader service
        :param level: The minimal log level of this handler
        """
        logging.Handler.__init__(self, level)
        self._framework = context.get_framework()
        self._reader = reader

    def _bundle_from_module(self, module_object):
        """
        Find the bundle associated to a module

        :param module_object: A Python module object
        :return: The Bundle object associated to the module, or None
        """
        try:
            # Get the module name
            module_object = module_object.__name__
        except AttributeError:
            # We got a string
            pass

        return self._framework.get_bundle_by_name(module_object)

    def emit(self, record):
        # pylint: disable=W0212
        """
        Handle a message logged with the logger

        :param record: A log record
        """
        # Get the bundle
        bundle = self._bundle_from_module(record.module)

        # Convert to a LogEntry
        entry = LogEntry(
            record.levelno, record.getMessage(), None, bundle, None
        )
        self._reader._store_entry(entry)

    def get_service(self, bundle, registration):
        # pylint: disable=W0613
        """
        Returns an instance of the log service for the given bundle

        :param bundle: Bundle consuming the service
        :param registration: Service registration bean
        :return: An instance of the logger
        """
        return LogServiceInstance(self._reader, bundle)

    @staticmethod
    def unget_service(bundle, registration):
        """
        Releases the service associated to the given bundle

        :param bundle: Consuming bundle
        :param registration: Service registration bean
        """
        pass


@BundleActivator
class Activator(object):
    """
    The bundle activator
    """

    def __init__(self):
        self.__reader_reg = None
        self.__factory_reg = None
        self.__factory = None

    @staticmethod
    def get_level(context):
        """
        Get the log level from the bundle context (framework properties)

        :param context: A bundle context
        :return: A log level (int)
        """
        # Get the log level
        level_value = context.get_property(PROPERTY_LOG_LEVEL)

        if level_value:
            for converter in int, logging.getLevelName:
                try:
                    parsed_level = converter(level_value)
                    if isinstance(parsed_level, int):
                        # Got a valid level
                        return parsed_level
                except (ValueError, TypeError):
                    pass

        # By default, use the INFO level
        return logging.INFO

    def start(self, context):
        """
        Bundle starting

        :param context: The bundle context
        """
        # Get the maximum number of entries authorized
        max_entries = context.get_property(PROPERTY_LOG_MAX_ENTRIES)
        try:
            # Normalize the value
            max_entries = int(max_entries)
        except (ValueError, TypeError):
            max_entries = 100

        # Register the LogReader service
        reader = LogReaderService(context, max_entries)
        self.__reader_reg = context.register_service(
            LOG_READER_SERVICE, reader, {}
        )

        # Register the LogService factory
        self.__factory = LogServiceFactory(
            context, reader, self.get_level(context)
        )
        self.__factory_reg = context.register_service(
            LOG_SERVICE, self.__factory, {}, factory=True
        )

        # Register the log service as a log handler
        logging.getLogger().addHandler(self.__factory)
        # ... but not for our own logs
        logger.removeHandler(self.__factory)

    def stop(self, _):
        """
        Bundle stopping
        """
        # Unregister the service
        if self.__factory_reg is not None:
            self.__factory_reg.unregister()
            self.__factory_reg = None

        if self.__reader_reg is not None:
            self.__reader_reg.unregister()
            self.__reader_reg = None

        # Unregister the handler
        logging.getLogger().removeHandler(self.__factory)
        self.__factory = None



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Dependency handler

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import logging
import threading

# Pelix beans
from pelix.constants import BundleActivator, BundleException
from pelix.internals.events import ServiceEvent

# iPOPO constants
import pelix.ipopo.constants as ipopo_constants
import pelix.ipopo.handlers.constants as constants

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class _HandlerFactory(constants.HandlerFactory):
    # pylint: disable=R0903
    """
    Factory service for service registration handlers
    """

    @staticmethod
    def _prepare_requirements(requirements, requires_filters):
        """
        Overrides the filters specified in the decorator with the given ones

        :param requirements: Dictionary of requirements (field → Requirement)
        :param requires_filters: Content of the 'requires.filter' component
                                 property (field → string)
        :return: The new requirements
        """
        if not requires_filters or not isinstance(requires_filters, dict):
            # No explicit filter configured
            return requirements

        # We need to change a part of the requirements
        new_requirements = {}
        for field, requirement in requirements.items():
            try:
                explicit_filter = requires_filters[field]

                # Store an updated copy of the requirement
                requirement_copy = requirement.copy()
                requirement_copy.set_filter(explicit_filter)
                new_requirements[field] = requirement_copy

            except (KeyError, TypeError, ValueError):
                # No information for this one, or invalid filter:
                # keep the factory requirement
                new_requirements[field] = requirement

        return new_requirements

    def get_handlers(self, component_context, instance):
        """
        Sets up service providers for the given component

        :param component_context: The ComponentContext bean
        :param instance: The component instance
        :return: The list of handlers associated to the given component
        """
        # Extract information from the context
        requirements = component_context.get_handler(
            ipopo_constants.HANDLER_REQUIRES
        )
        requires_filters = component_context.properties.get(
            ipopo_constants.IPOPO_REQUIRES_FILTERS, None
        )

        # Prepare requirements
        requirements = self._prepare_requirements(
            requirements, requires_filters
        )

        # Set up the runtime dependency handlers
        handlers = []
        for field, requirement in requirements.items():
            # Construct the handler
            if requirement.aggregate:
                handlers.append(AggregateDependency(field, requirement))
            else:
                handlers.append(SimpleDependency(field, requirement))

        return handlers


@BundleActivator
class _Activator(object):
    """
    The bundle activator
    """

    def __init__(self):
        """
        Sets up members
        """
        self._registration = None

    def start(self, context):
        """
        Bundle started
        """
        # Set up properties
        properties = {
            constants.PROP_HANDLER_ID: ipopo_constants.HANDLER_REQUIRES
        }

        # Register the handler factory service
        self._registration = context.register_service(
            constants.SERVICE_IPOPO_HANDLER_FACTORY,
            _HandlerFactory(),
            properties,
        )

    def stop(self, _):
        """
        Bundle stopped
        """
        # Unregister the service
        self._registration.unregister()
        self._registration = None


# ------------------------------------------------------------------------------


class _RuntimeDependency(constants.DependencyHandler):
    """
    Manages a required dependency field when a component is running
    """

    def __init__(self, field, requirement):
        """
        Sets up the dependency

        :param field: The injected field name
        :param requirement: The Requirement describing this dependency
        """
        # The internal state lock
        self._lock = threading.RLock()

        # The iPOPO StoredInstance object (given during manipulation)
        self._ipopo_instance = None

        # The bundle context
        self._context = None

        # The associated field
        self._field = field

        # The underlying requirement
        self.requirement = requirement

        # Current field value
        self._value = None

    def manipulate(self, stored_instance, component_instance):
        """
        Stores the given StoredInstance bean.

        :param stored_instance: The iPOPO component StoredInstance
        :param component_instance: The component instance
        """
        # Store the stored instance...
        self._ipopo_instance = stored_instance

        # ... and the bundle context
        self._context = stored_instance.bundle_context

    def clear(self):
        """
        Cleans up the manager. The manager can't be used after this method has
        been called
        """
        self._lock = None
        self._ipopo_instance = None
        self._context = None
        self.requirement = None
        self._value = None
        self._field = None

    def get_bindings(self):
        """
        Retrieves the list of the references to the bound services

        :return: A list of ServiceReferences objects
        """
        raise NotImplementedError

    def get_field(self):
        """
        Returns the name of the field handled by this handler
        """
        return self._field

    def get_kinds(self):
        """
        Retrieves the kinds of this handler: 'dependency'

        :return: the kinds of this handler
        """
        return (constants.KIND_DEPENDENCY,)

    def get_value(self):
        """
        Retrieves the value to inject in the component (can be different of
        self._value)

        :return: The value to inject
        """
        return self._value

    def is_valid(self):
        """
        Tests if the dependency is in a valid state
        """
        return (
            self.requirement is not None and self.requirement.optional
        ) or self._value is not None

    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        """
        raise NotImplementedError

    def on_service_departure(self, svc_ref):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        """
        raise NotImplementedError

    def on_service_modify(self, svc_ref, old_properties):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        :param old_properties: Previous properties values
        """
        raise NotImplementedError

    def service_changed(self, event):
        """
        Called by the framework when a service event occurs
        """
        if (
            self._ipopo_instance is None
            or not self._ipopo_instance.check_event(event)
        ):
            # stop() and clean() may have been called after we have been put
            # inside a listener list copy...
            # or we've been told to ignore this event
            return

        # Call sub-methods
        kind = event.get_kind()
        svc_ref = event.get_service_reference()

        if kind == ServiceEvent.REGISTERED:
            # Service coming
            self.on_service_arrival(svc_ref)

        elif kind in (
            ServiceEvent.UNREGISTERING,
            ServiceEvent.MODIFIED_ENDMATCH,
        ):
            # Service gone or not matching anymore
            self.on_service_departure(svc_ref)

        elif kind == ServiceEvent.MODIFIED:
            # Modified properties (can be a new injection)
            self.on_service_modify(svc_ref, event.get_previous_properties())

    def start(self):
        """
        Starts the dependency manager
        """
        self._context.add_service_listener(
            self, self.requirement.filter, self.requirement.specification
        )

    def stop(self):
        """
        Stops the dependency manager (must be called before clear())

        :return: The removed bindings (list) or None
        """
        self._context.remove_service_listener(self)


class SimpleDependency(_RuntimeDependency):
    """
    Manages a simple dependency field
    """

    def __init__(self, field, requirement):
        """
        Sets up the dependency
        """
        super(SimpleDependency, self).__init__(field, requirement)

        # We have only one reference to keep
        self.reference = None

        # Pending reference (to avoid double-lookup)
        self._pending_ref = None

    def clear(self):
        """
        Cleans up the manager. The manager can't be used after this method has
        been called
        """
        self.reference = None
        self._pending_ref = None
        super(SimpleDependency, self).clear()

    def get_bindings(self):
        """
        Retrieves the list of the references to the bound services

        :return: A list of ServiceReferences objects
        """
        with self._lock:
            if self.reference:
                return [self.reference]

            return []

    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        """
        with self._lock:
            if self._value is None:
                # Inject the service
                self.reference = svc_ref
                self._value = self._context.get_service(svc_ref)

                self._ipopo_instance.bind(self, self._value, self.reference)
                return True

        return None

    def on_service_departure(self, svc_ref):
        """
        Called when a service has been unregistered from the framework

        :param svc_ref: A service reference
        """
        with self._lock:
            if svc_ref is self.reference:
                service = self._value

                # Clear the instance values
                self._value = None
                self.reference = None

                if self.requirement.immediate_rebind:
                    # Look for a replacement
                    self._pending_ref = self._context.get_service_reference(
                        self.requirement.specification, self.requirement.filter
                    )

                self._ipopo_instance.unbind(self, service, svc_ref)
                return True

            return None

    def on_service_modify(self, svc_ref, old_properties):
        """
        Called when a service has been modified in the framework

        :param svc_ref: A service reference
        :param old_properties: Previous properties values
        """
        with self._lock:
            if self.reference is None:
                # A previously registered service now matches our filter
                self.on_service_arrival(svc_ref)
            elif svc_ref is self.reference:
                # Notify the property modification
                self._ipopo_instance.update(
                    self, self._value, svc_ref, old_properties
                )

    def stop(self):
        """
        Stops the dependency manager (must be called before clear())

        :return: The removed bindings (list) or None
        """
        super(SimpleDependency, self).stop()
        if self.reference is not None:
            # Return a tuple of tuple
            return ((self._value, self.reference),)

        return None

    def is_valid(self):
        """
        Tests if the dependency is in a valid state
        """
        return super(SimpleDependency, self).is_valid() or (
            self.requirement.immediate_rebind and self._pending_ref is not None
        )

    def try_binding(self):
        """
        Searches for the required service if needed

        :raise BundleException: Invalid ServiceReference found
        """
        with self._lock:
            if self.reference is not None:
                # Already bound
                return

            if self._pending_ref is not None:
                # Get the reference we chose to keep this component valid
                ref = self._pending_ref
                self._pending_ref = None
            else:
                # Get the first matching service
                ref = self._context.get_service_reference(
                    self.requirement.specification, self.requirement.filter
                )

            if ref is not None:
                # Found a service
                self.on_service_arrival(ref)


class AggregateDependency(_RuntimeDependency):
    """
    Manages an aggregated dependency field
    """

    def __init__(self, field, requirement):
        """
        Sets up the dependency
        """
        super(AggregateDependency, self).__init__(field, requirement)

        # Reference -> Service
        self.services = {}

        # Future injected value
        self._future_value = None

    def clear(self):
        """
        Cleans up the manager. The manager can't be used after this method has
        been called
        """
        self.services.clear()
        self.services = None
        self._future_value = None
        super(AggregateDependency, self).clear()

    def get_bindings(self):
        """
        Retrieves the list of the references to the bound services

        :return: A list of ServiceReferences objects
        """
        with self._lock:
            return list(self.services.keys())

    def get_value(self):
        """
        Retrieves the value to inject in the component

        :return: The value to inject
        """
        with self._lock:
            # The value field must be a copy of our list
            if self._future_value is not None:
                return self._future_value[:]

            return None

    def is_valid(self):
        """
        Tests if the dependency is in a valid state
        """
        return (
            self.requirement is not None and self.requirement.optional
        ) or self._future_value is not None

    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        """
        with self._lock:
            if svc_ref not in self.services:
                # Get the new service
                service = self._context.get_service(svc_ref)

                if self._future_value is None:
                    # First value
                    self._future_value = []

                # Store the information
                self._future_value.append(service)
                self.services[svc_ref] = service

                self._ipopo_instance.bind(self, service, svc_ref)
                return True

            return None

    def on_service_departure(self, svc_ref):
        """
        Called when a service has been unregistered from the framework

        :param svc_ref: A service reference
        :return: A tuple (service, reference) if the service has been lost,
                 else None
        """
        with self._lock:
            try:
                # Get the service instance
                service = self.services.pop(svc_ref)
            except KeyError:
                # Not a known service reference: ignore
                pass
            else:
                # Clean the instance values
                self._future_value.remove(service)

                # Nullify the value if needed
                if not self._future_value:
                    self._future_value = None

                self._ipopo_instance.unbind(self, service, svc_ref)
                return True

            return None

    def on_service_modify(self, svc_ref, old_properties):
        """
        Called when a service has been modified in the framework

        :param svc_ref: A service reference
        :param old_properties: Previous properties values
        :return: A tuple (added, (service, reference)) if the dependency has
                 been changed, else None
        """
        with self._lock:
            try:
                # Look for the service
                service = self.services[svc_ref]
            except KeyError:
                # A previously registered service now matches our filter
                return self.on_service_arrival(svc_ref)
            else:
                # Notify the property modification
                self._ipopo_instance.update(
                    self, service, svc_ref, old_properties
                )

            return None

    def stop(self):
        """
        Stops the dependency manager (must be called before clear())

        :return: The removed bindings (list) or None
        """
        super(AggregateDependency, self).stop()

        if self.services:
            return [
                (service, reference)
                for reference, service in self.services.items()
            ]

        return None

    def try_binding(self):
        """
        Searches for the required service if needed

        :raise BundleException: Invalid ServiceReference found
        """
        with self._lock:
            if self.services:
                # We already are alive (not our first call)
                # => we are updated through service events
                return

            # Get all matching services
            refs = self._context.get_all_service_references(
                self.requirement.specification, self.requirement.filter
            )
            if not refs:
                # No match found
                return

            results = []
            try:
                # Bind all new reference
                for reference in refs:
                    added = self.on_service_arrival(reference)
                    if added:
                        results.append(reference)
            except BundleException as ex:
                # Get the logger for this instance
                logger = logging.getLogger(
                    "-".join((self._ipopo_instance.name, "AggregateDependency"))
                )
                logger.debug("Error binding multiple references: %s", ex)

                # Undo what has just been done, ignoring errors
                for reference in results:
                    try:
                        self.on_service_departure(reference)
                    except BundleException as ex2:
                        logger.debug("Error cleaning up: %s", ex2)

                del results[:]
                raise



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
iPOPO remote shell

Provides a remote interface for the Pelix shell that can be accessed using
telnet or netcat.

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import print_function

# Standard library
from select import select
import argparse
import logging
import threading
import socket
import sys

try:
    # Some Python distributions don't support SSL
    import ssl
except ImportError:
    ssl = None

try:
    # Python 3
    # pylint: disable=F0401
    import socketserver
except ImportError:
    # Python 2
    # pylint: disable=F0401
    import SocketServer as socketserver

# iPOPO
from pelix.ipopo.constants import use_ipopo
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Property,
    Validate,
    Invalidate,
    Provides,
    HiddenProperty,
)

# Pelix
from pelix.shell.console import make_common_parser, handle_common_arguments
import pelix.framework
import pelix.ipv6utils
import pelix.shell
import pelix.shell.beans as beans
import pelix.utilities as utilities

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------


class SharedBoolean(object):
    """
    Shared boolean between objects / threads
    """

    def __init__(self, value=False):
        """
        Set up members
        """
        self._lock = threading.Lock()
        self._value = value

    def get_value(self):
        """
        Retrieves the boolean value
        """
        with self._lock:
            return self._value

    def set_value(self, value):
        """
        Sets the boolean value
        """
        with self._lock:
            self._value = value


# ------------------------------------------------------------------------------


class RemoteConsole(socketserver.StreamRequestHandler):
    """
    Handles incoming connections and redirect network stream to the Pelix shell
    """

    def __init__(self, shell_svc, active_flag, *args):
        """
        Sets up members

        :param shell_svc: The underlying Pelix shell service
        :param active_flag: Common flag for stopping the client communication
        """
        self._shell = shell_svc
        self._active = active_flag
        socketserver.StreamRequestHandler.__init__(self, *args)

    def send(self, data):
        """
        Tries to send data to the client.

        :param data: Data to be sent
        :return: True if the data was sent, False on error
        """
        if data is not None:
            data = data.encode("UTF-8")

        try:
            self.wfile.write(data)
            self.wfile.flush()
            return True

        except IOError:
            # An error occurred, mask it
            # -> This allows to handle the command even if the client has been
            # disconnect (i.e. "echo stop 0 | nc localhost 9000")
            return False

    def handle(self):
        """
        Handles a TCP client
        """
        _logger.info(
            "RemoteConsole client connected: [%s]:%d",
            self.client_address[0],
            self.client_address[1],
        )

        # Prepare the session
        session = beans.ShellSession(
            beans.IOHandler(self.rfile, self.wfile),
            {"remote_client_ip": self.client_address[0]},
        )

        # Print the banner
        def get_ps1():
            """
            Gets the prompt string from the session of the shell service

            :return: The prompt string
            """
            try:
                return session.get("PS1")
            except KeyError:
                return self._shell.get_ps1()

        self.send(self._shell.get_banner())
        self.send(get_ps1())

        try:
            while self._active.get_value():
                # Wait for data
                rlist = select([self.connection], [], [], .5)[0]
                if not rlist:
                    # Nothing to do (poll timed out)
                    continue

                data = self.rfile.readline()
                if not data:
                    # End of stream (client gone)
                    break

                # Strip the line
                line = data.strip()
                if not data:
                    # Empty line
                    continue

                # Execute it
                try:
                    self._shell.handle_line(line, session)
                except KeyboardInterrupt:
                    # Stop there on interruption
                    self.send("\nInterruption received.")
                    return
                except IOError as ex:
                    # I/O errors are fatal
                    _logger.exception(
                        "Error communicating with a client: %s", ex
                    )
                    break
                except Exception as ex:
                    # Other exceptions are not important
                    import traceback

                    self.send("\nError during last command: {0}\n".format(ex))
                    self.send(traceback.format_exc())

                # Print the prompt
                self.send(get_ps1())
        finally:
            _logger.info(
                "RemoteConsole client gone: [%s]:%d",
                self.client_address[0],
                self.client_address[1],
            )

            try:
                # Be polite
                self.send("\nSession closed. Good bye.\n")
                self.finish()
            except IOError as ex:
                _logger.warning("Error cleaning up connection: %s", ex)


# ------------------------------------------------------------------------------


class ThreadingTCPServerFamily(socketserver.ThreadingTCPServer):
    """
    Threaded TCP Server handling different address families
    """

    def __init__(
        self,
        server_address,
        request_handler_class,
        cert_file=None,
        key_file=None,
        key_password=None,
        ca_file=None,
    ):
        """
        Sets up the TCP server. Doesn't bind nor activate it.

        :param server_address: Server binding address
        :param request_handler_class: Class to instantiate for each client
        :param cert_file: Path to the server certificate
        :param key_file: Path to the server private key
        :param key_password: Password for the key file
        :param ca_file: Path to Certificate Authority to authenticate clients
        """
        # Determine the address family
        addr_info = socket.getaddrinfo(
            server_address[0], server_address[1], 0, 0, socket.SOL_TCP
        )

        # Change the address family before the socket is created
        # Get the family of the first possibility
        self.address_family = addr_info[0][0]

        # Keep track of SSL arguments
        self.cert_file = cert_file
        self.key_file = key_file
        self.key_password = key_password
        self.ca_file = ca_file

        # Call the super constructor
        socketserver.ThreadingTCPServer.__init__(
            self, server_address, request_handler_class, False
        )
        if self.address_family == socket.AF_INET6:
            # Explicitly ask to be accessible both by IPv4 and IPv6
            try:
                pelix.ipv6utils.set_double_stack(self.socket)
            except AttributeError as ex:
                _logger.exception("System misses IPv6 constant: %s", ex)
            except socket.error as ex:
                _logger.exception("Error setting up IPv6 double stack: %s", ex)

    def get_request(self):
        """
        Accepts a new client. Sets up SSL wrapping if necessary.

        :return: A tuple: (client socket, client address tuple)
        """
        # Accept the client
        client_socket, client_address = self.socket.accept()

        if ssl is not None and self.cert_file:
            # Setup an SSL context to accept clients with a certificate
            # signed by a known chain of authority.
            # Other clients will be rejected during handshake.
            context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            try:
                # Force a valid/signed client-side certificate
                context.verify_mode = ssl.CERT_REQUIRED

                # Load the server certificate
                context.load_cert_chain(
                    certfile=self.cert_file,
                    keyfile=self.key_file,
                    password=self.key_password,
                )

                if self.ca_file:
                    # Load the given authority chain
                    context.load_verify_locations(self.ca_file)
                else:
                    # Load the default chain if none given
                    context.load_default_certs(ssl.Purpose.CLIENT_AUTH)
            except Exception as ex:
                # Explicitly log the error as the default behaviour hides it
                _logger.error("Error setting up the SSL context: %s", ex)
                raise

            try:
                # SSL handshake
                client_stream = context.wrap_socket(
                    client_socket, server_side=True
                )
            except ssl.SSLError as ex:
                # Explicitly log the exception before re-raising it
                _logger.warning(
                    "Error during SSL handshake with %s: %s", client_address, ex
                )
                raise
        else:
            # Nothing to do, use the raw socket
            client_stream = client_socket

        return client_stream, client_address

    def process_request(self, request, client_address):
        """
        Starts a new thread to process the request, adding the client address
        in its name.
        """
        thread = threading.Thread(
            name="RemoteShell-{0}-Client-{1}".format(
                self.server_address[1], client_address[:2]
            ),
            target=self.process_request_thread,
            args=(request, client_address),
        )
        thread.daemon = self.daemon_threads
        thread.start()


def _create_server(
    shell,
    server_address,
    port,
    cert_file=None,
    key_file=None,
    key_password=None,
    ca_file=None,
):
    """
    Creates the TCP console on the given address and port

    :param shell: The remote shell handler
    :param server_address: Server bound address
    :param port: Server port
    :param cert_file: Path to the server certificate
    :param key_file: Path to the server private key
    :param key_password: Password for the key file
    :param ca_file: Path to Certificate Authority to authenticate clients
    :return: A tuple: Server thread, TCP server object, Server active flag
    """
    # Set up the request handler creator
    active_flag = SharedBoolean(True)

    def request_handler(*rh_args):
        """
        Constructs a RemoteConsole as TCP request handler
        """
        return RemoteConsole(shell, active_flag, *rh_args)

    # Set up the server
    server = ThreadingTCPServerFamily(
        (server_address, port),
        request_handler,
        cert_file,
        key_file,
        key_password,
        ca_file,
    )

    # Set flags
    server.daemon_threads = True
    server.allow_reuse_address = True

    # Activate the server
    server.server_bind()
    server.server_activate()

    # Serve clients
    server_thread = threading.Thread(
        target=server.serve_forever, name="RemoteShell-{0}".format(port)
    )
    server_thread.daemon = True
    server_thread.start()

    return server_thread, server, active_flag


# ------------------------------------------------------------------------------


@ComponentFactory(pelix.shell.FACTORY_REMOTE_SHELL)
@Provides(pelix.shell.SERVICE_SHELL_REMOTE)
@Requires("_shell", pelix.shell.SERVICE_SHELL)
@Property("_address", "pelix.shell.address", "localhost")
@Property("_port", "pelix.shell.port", 9000)
@Property("_ca_file", "pelix.shell.ssl.ca", None)
@Property("_cert_file", "pelix.shell.ssl.cert", None)
@Property("_key_file", "pelix.shell.ssl.key", None)
@HiddenProperty("_key_password", "pelix.shell.ssl.key_password", None)
class IPopoRemoteShell(object):
    """
    The iPOPO Remote Shell, based on the Pelix Shell
    """

    def __init__(self):
        """
        Sets up the component
        """
        # Component shell
        self._shell = None
        self._address = None
        self._port = 0

        # SSL configuration
        self._ca_file = None
        self._cert_file = None
        self._key_file = None
        self._key_password = None

        # Internals
        self._thread = None
        self._server = None
        self._server_flag = None

    def get_access(self):
        """
        Implementation of the remote shell specification

        :return: A (host, port) tuple
        """
        return self._address, self._port

    def get_banner(self):
        """
        Retrieves the shell banner

        :return: The shell banner
        """
        line = "-" * 72
        shell_banner = self._shell.get_banner()

        return "{lines}\n{shell_banner}\niPOPO Remote Shell\n{lines}\n".format(
            lines=line, shell_banner=shell_banner
        )

    def get_ps1(self):
        """
        Returns the shell prompt

        :return: The shell prompt
        """
        return self._shell.get_ps1()

    def handle_line(self, line, session):
        """
        Handles the command line.

        **Does not catch exceptions !**

        :param line: The command line
        :param session: The current shell session
        :return: The execution result (True on success, else False)
        """
        return self._shell.execute(line, session)

    @Validate
    def validate(self, _):
        """
        Component validation
        """
        if not self._address:
            # Local host by default
            self._address = "localhost"

        try:
            self._port = int(self._port)
            if self._port < 0 or self._port > 65535:
                # Invalid port value
                self._port = 0
        except (ValueError, TypeError):
            # Invalid port string: use a random port
            self._port = 0

        # Start the TCP server
        self._thread, self._server, self._server_flag = _create_server(
            self,
            self._address,
            self._port,
            self._cert_file,
            self._key_file,
            self._key_password,
            self._ca_file,
        )

        # Property update (if port was 0)
        self._port = self._server.socket.getsockname()[1]
        _logger.info("RemoteShell validated on port: %d", self._port)

    @Invalidate
    def invalidate(self, _):
        """
        Component invalidation
        """
        # Stop the clients loops
        if self._server is not None:
            self._server_flag.set_value(False)

            # Shutdown the server
            self._server.shutdown()
            self._thread.join(2)

            # Close the server socket (ignore errors)
            self._server.server_close()
            _logger.info("RemoteShell gone from port: %d", self._port)

        # Clean up
        self._thread = None
        self._server = None
        self._server_flag = None


# ------------------------------------------------------------------------------


def _run_interpreter(variables, banner):
    """
    Runs a Python interpreter console and blocks until the user exits it.

    :param variables: Interpreters variables (locals)
    :param banner: Start-up banners
    """
    # Script-only imports
    import code

    try:
        import readline
        import rlcompleter

        readline.set_completer(rlcompleter.Completer(variables).complete)
        readline.parse_and_bind("tab: complete")
    except ImportError:
        # readline is not available: ignore
        pass

    # Start the console
    shell = code.InteractiveConsole(variables)
    shell.interact(banner)


def main(argv=None):
    """
    Script entry point

    :param argv: Script arguments (None for sys.argv)
    :return: An exit code or None
    """
    # Prepare arguments
    parser = argparse.ArgumentParser(
        prog="pelix.shell.remote",
        parents=[make_common_parser()],
        description="Pelix Remote Shell ({} SSL support)".format(
            "with" if ssl is not None else "without"
        ),
    )

    # Remote shell options
    group = parser.add_argument_group("Remote Shell options")
    group.add_argument(
        "-a",
        "--address",
        default="localhost",
        help="The remote shell binding address",
    )
    group.add_argument(
        "-p",
        "--port",
        type=int,
        default=9000,
        help="The remote shell binding port",
    )

    if ssl is not None:
        # Remote Shell TLS options
        group = parser.add_argument_group("TLS Options")
        group.add_argument("--cert", help="Path to the server certificate file")
        group.add_argument(
            "--key",
            help="Path to the server key file "
            "(can be omitted if the key is in the certificate)",
        )
        group.add_argument(
            "--key-password",
            help="Password of the server key."
            "Set to '-' for a password request.",
        )
        group.add_argument(
            "--ca-chain",
            help="Path to the CA chain file to authenticate clients",
        )

    # Local options
    group = parser.add_argument_group("Local options")
    group.add_argument(
        "--no-input",
        action="store_true",
        help="Run without input (for daemon mode)",
    )

    # Parse them
    args = parser.parse_args(argv)

    # Handle arguments
    init = handle_common_arguments(args)

    # Set the initial bundles
    bundles = [
        "pelix.ipopo.core",
        "pelix.shell.core",
        "pelix.shell.ipopo",
        "pelix.shell.remote",
    ]
    bundles.extend(init.bundles)

    # Start a Pelix framework
    framework = pelix.framework.create_framework(
        utilities.remove_duplicates(bundles), init.properties
    )
    framework.start()
    context = framework.get_bundle_context()

    # Instantiate configured components
    init.instantiate_components(framework.get_bundle_context())

    # Instantiate a Remote Shell, if necessary
    with use_ipopo(context) as ipopo:
        rshell_name = "remote-shell"
        try:
            ipopo.get_instance_details(rshell_name)
        except ValueError:
            # Component doesn't exist, we can instantiate it.

            if ssl is not None:
                # Copy parsed arguments
                ca_chain = args.ca_chain
                cert = args.cert
                key = args.key

                # Normalize the TLS key file password argument
                if args.key_password == "-":
                    import getpass

                    key_password = getpass.getpass(
                        "Password for {}: ".format(args.key or args.cert)
                    )
                else:
                    key_password = args.key_password
            else:
                # SSL support is missing:
                # Ensure the SSL arguments are defined but set to None
                ca_chain = None
                cert = None
                key = None
                key_password = None

            # Setup the component
            rshell = ipopo.instantiate(
                pelix.shell.FACTORY_REMOTE_SHELL,
                rshell_name,
                {
                    "pelix.shell.address": args.address,
                    "pelix.shell.port": args.port,
                    "pelix.shell.ssl.ca": ca_chain,
                    "pelix.shell.ssl.cert": cert,
                    "pelix.shell.ssl.key": key,
                    "pelix.shell.ssl.key_password": key_password,
                },
            )

            # Avoid loose reference to the password
            del key_password
        else:
            logging.error(
                "A remote shell component (%s) is already "
                "configured. Abandon.",
                rshell_name,
            )
            return 1

    # Prepare a banner
    host, port = rshell.get_access()
    try:
        if args.no_input:
            # No input required: just print the access to the shell
            print("Remote shell bound to:", host, "- port:", port)

            try:
                while not framework.wait_for_stop(1):
                    # Awake from wait every second to let KeyboardInterrupt
                    # exception to raise
                    pass
            except KeyboardInterrupt:
                print("Got Ctrl+C: exiting.")
                return 127
        else:
            # Prepare interpreter variables
            variables = {
                "__name__": "__console__",
                "__doc__": None,
                "__package__": None,
                "framework": framework,
                "context": context,
                "use_ipopo": use_ipopo,
            }

            banner = (
                "{lines}\nPython interpreter with Pelix Remote Shell\n"
                "Remote shell bound to: {host}:{port}\n{lines}\n"
                "Python version: {version}\n".format(
                    lines="-" * 80, version=sys.version, host=host, port=port
                )
            )

            # Run an interpreter
            _run_interpreter(variables, banner)
    finally:
        # Stop the framework
        framework.stop()


if __name__ == "__main__":
    # Run the entry point
    sys.exit(main() or 0)



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""

Py4j-based Distribution and Discovery Provider

:author: Scott Lewis
:copyright: Copyright 2020, Scott Lewis
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Scott Lewis

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from threading import Thread, RLock
import logging

from osgiservicebridge.bridge import (
    JavaServiceProxy,
    Py4jServiceBridgeEventListener,
    Py4jServiceBridge,
    PythonService,
)
from osgiservicebridge.protobuf import (
    ProtobufJavaServiceProxy,
    ProtobufPythonService,
)

from py4j.java_gateway import GatewayParameters, CallbackServerParameters
from py4j.java_gateway import DEFAULT_PORT, DEFAULT_PYTHON_PROXY_PORT

# needed ipopo decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Provides,
    Instantiate,
    Property,
    Validate,
    ValidateComponent,
    Invalidate,
    PostRegistration,
)

from pelix.ipopo.constants import (
    ARG_BUNDLE_CONTEXT,
    ARG_PROPERTIES,
)

# Providers API
from pelix.rsa import prop_dot_suffix
from pelix.rsa.providers.distribution import (
    Container,
    ExportContainer,
    ImportContainer,
    DistributionProvider,
    SERVICE_EXPORT_CONTAINER,
    SERVICE_IMPORT_CONTAINER,
    SERVICE_EXPORT_DISTRIBUTION_PROVIDER,
    SERVICE_IMPORT_DISTRIBUTION_PROVIDER,
)
from pelix.rsa.endpointdescription import EndpointDescription

# ------------------------------------------------------------------------------
# Module version

__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

_logger = logging.getLogger(__name__)

# ------------------------------------------------------------------------------
# Note:  These must match the Java-side constants recored in Java interface
# class: org.eclipse.ecf.provider.py4j.Py4jConstants
ECF_PY4J_CONTAINER_CONFIG_TYPE = "ecf.py4j"
ECF_PY4J_NAMESPACE = "ecf.namespace.py4j"

ECF_PY4J_JAVA_HOST_CONFIG_TYPE = "ecf.py4j.host"
ECF_PY4J_JAVA_CONSUMER_CONFIG_TYPE = "ecf.py4j.consumer"
ECF_PY4J_PYTHON_HOST_CONFIG_TYPE = "ecf.py4j.host.python"
ECF_PY4J_PYTHON_CONSUMER_CONFIG_TYPE = "ecf.py4j.consumer.python"
ECF_PY4J_SUPPORTED_INTENTS = [
    "exactlyOnce",
    "passByReference",
    "ordered",
    "py4j",
    "py4j.async",
    "osgi.async",
    "osgi.private",
]
# Protobuf
ECF_PY4JPB_JAVA_HOST_CONFIG_TYPE = "ecf.py4j.protobuf.host"
ECF_PY4JPB_JAVA_CONSUMER_CONFIG_TYPE = "ecf.py4j.protobuf.consumer"
ECF_PY4JPB_PYTHON_HOST_CONFIG_TYPE = "ecf.py4j.python.protobuf.host"
ECF_PY4JPB_PYTHON_CONSUMER_CONFIG_TYPE = "ecf.py4j.python.protobuf.consumer"
ECF_PY4JPB_SUPPORTED_INTENTS = [
    "exactlyOnce",
    "passByReference",
    "passByValue",
    "ordered",
    "py4j",
    "py4j.protobuf",
    "py4j.async",
    "osgi.async",
    "osgi.private",
]

ECF_PY4J_JAVA_PORT_PROP = "javaport"
ECF_PY4J_PYTHON_PORT_PROP = "pythonport"
ECF_PY4J_DEFAULT_SERVICE_TIMEOUT = "defaultservicetimeout"

# ------------------------------------------------------------------------------


@ComponentFactory(ECF_PY4J_CONTAINER_CONFIG_TYPE)
@Provides([SERVICE_EXPORT_CONTAINER, SERVICE_IMPORT_CONTAINER])
class Py4jContainer(ExportContainer, ImportContainer):
    def __init__(self, max_workers=5):
        ExportContainer.__init__(self)
        ImportContainer.__init__(self)
        self._max_workers = max_workers
        self._executor = None

    @ValidateComponent(ARG_BUNDLE_CONTEXT, ARG_PROPERTIES)
    def _validate_component(self, bundle_context, container_props):
        Container._validate_component(self, bundle_context, container_props)
        self._executor = ThreadPoolExecutor(max_workers=self._max_workers)

    @Invalidate
    def _invalidate_component(self, _):
        Container._invalidate_component(self, _)
        if self._executor:
            self._executor.shutdown()
            self._executor = None

    def get_connected_id(self):
        return ExportContainer.get_connected_id(self)

    def _export_service(self, svc, ed):
        # pylint: disable=W0212
        # modify svc class to have appropriate metadata for py4j
        timeout = ed.get_osgi_basic_timeout()
        if not timeout:
            timeout = 30

        args = [
            self._get_distribution_provider()._get_bridge(),
            ed.get_interfaces(),
            svc,
            self._executor,
            timeout,
        ]

        if (
            ECF_PY4JPB_PYTHON_HOST_CONFIG_TYPE
            in ed.get_remote_configs_supported()
        ):
            clazz = ProtobufPythonService
        else:
            clazz = PythonService

        psvc = clazz(*args)

        self._get_distribution_provider()._get_bridge().export(
            psvc, ed.get_properties()
        )
        ExportContainer._export_service(self, psvc, ed)
        return True

    def _unexport_service(self, ed):
        # pylint: disable=W0212
        dp = self._get_distribution_provider()
        if dp:
            bridge = dp._get_bridge()
            if bridge:
                bridge.unexport(ed.get_id())
        ExportContainer._unexport_service(self, ed)
        return True

    def _prepare_proxy(self, endpoint_description):
        # pylint: disable=W0212
        # lookup the bridge proxy associated with the
        # endpoint_description.get_id()
        bridge = self._get_distribution_provider()._get_bridge()
        proxy = bridge.get_import_endpoint(endpoint_description.get_id())[0]
        timeout = endpoint_description.get_osgi_basic_timeout()
        if not timeout:
            timeout = self._container_props.get(
                ECF_PY4J_DEFAULT_SERVICE_TIMEOUT, 30
            )

        args = [
            bridge.get_jvm(),
            endpoint_description.get_interfaces(),
            proxy,
            self._executor,
            timeout,
        ]

        clazz = JavaServiceProxy

        if (
            ECF_PY4JPB_JAVA_HOST_CONFIG_TYPE
            in endpoint_description.get_remote_configs_supported()
        ):
            clazz = ProtobufJavaServiceProxy

        return clazz(*args)

    def unimport_service(self, endpoint_description):
        # pylint: disable=W0212
        dp = self._get_distribution_provider()
        if dp:
            bridge = dp._get_bridge()
            if bridge:
                bridge.remove_import_endpoint(endpoint_description.get_id())
        ImportContainer.unimport_service(self, endpoint_description)


@ComponentFactory("py4j-distribution-provider-factory")
@Provides(
    [SERVICE_EXPORT_DISTRIBUTION_PROVIDER, SERVICE_IMPORT_DISTRIBUTION_PROVIDER]
)
@Property("_config_name", "config_name", ECF_PY4J_CONTAINER_CONFIG_TYPE)
@Property("_namespace", "namespace", ECF_PY4J_NAMESPACE)
@Property(
    "_supported_configs",
    "supported_configs",
    [ECF_PY4J_PYTHON_HOST_CONFIG_TYPE, ECF_PY4J_PYTHON_CONSUMER_CONFIG_TYPE],
)
@Property("_supported_intents", "supported_intents",
          ECF_PY4J_SUPPORTED_INTENTS)
@Property(
    "_supported_pb_intents",
    "supported_pb_intents",
    ECF_PY4JPB_SUPPORTED_INTENTS,
)
@Property(
    "_javaport",
    prop_dot_suffix(ECF_PY4J_CONTAINER_CONFIG_TYPE, ECF_PY4J_JAVA_PORT_PROP),
    DEFAULT_PORT,
)
@Property(
    "_pythonport",
    prop_dot_suffix(ECF_PY4J_CONTAINER_CONFIG_TYPE, ECF_PY4J_PYTHON_PORT_PROP),
    DEFAULT_PYTHON_PROXY_PORT,
)
@Property(
    "_default_service_timeout",
    prop_dot_suffix(
        ECF_PY4J_CONTAINER_CONFIG_TYPE, ECF_PY4J_DEFAULT_SERVICE_TIMEOUT
    ),
    30,
)
@Instantiate("py4j-distribution-provider")
class Py4jDistributionProvider(
    DistributionProvider, Py4jServiceBridgeEventListener
):
    def __init__(self):
        super(Py4jDistributionProvider, self).__init__()
        self._bridge = None
        self._container = None
        self._queue = Queue()
        self._thread = Thread(target=self._worker)
        self._thread.daemon = True
        self._done = False
        self._lock = RLock()
        self._py4jcontainer = self._supported_pb_intents = None
        self._javaport = self._pythonport = self._default_service_timeout = None

    def _get_bridge(self):
        return self._bridge

    # Override of DistributionProvider._get_imported_configs. Returns
    # the Py4j bridge.get_id() in list
    def _get_imported_configs(self, exported_configs):
        imported_configs = []
        if ECF_PY4JPB_JAVA_HOST_CONFIG_TYPE in exported_configs:
            imported_configs.append(ECF_PY4JPB_PYTHON_HOST_CONFIG_TYPE)
        if ECF_PY4J_JAVA_HOST_CONFIG_TYPE in exported_configs:
            imported_configs.append(ECF_PY4J_PYTHON_HOST_CONFIG_TYPE)
        return imported_configs

    # Implementation of ImportDistributionProvider
    def supports_import(self, exported_configs, service_intents, import_props):
        # pylint: disable=W0613
        if ECF_PY4JPB_JAVA_HOST_CONFIG_TYPE in exported_configs:
            if self._match_intents_supported(
                service_intents, self._supported_pb_intents
            ):
                return self._container
        elif ECF_PY4J_JAVA_HOST_CONFIG_TYPE in exported_configs:
            if self._match_intents(service_intents):
                return self._container

        return None

    # Implementation of ExportDistributionProvider
    def supports_export(self, exported_configs, service_intents, export_props):
        # pylint: disable=W0613
        if self._match_intents(service_intents):
            if (
                ECF_PY4J_PYTHON_HOST_CONFIG_TYPE in exported_configs
                or ECF_PY4JPB_PYTHON_HOST_CONFIG_TYPE in exported_configs
            ):
                return self._container

        return None

    @Validate
    def _validate(self, _):
        # here is where we can get java and python ports and change the
        # defaults for connecting
        try:
            self._bridge = Py4jServiceBridge(
                service_listener=self,
                gateway_parameters=GatewayParameters(port=self._javaport),
                callback_server_parameters=CallbackServerParameters(
                    port=self._pythonport
                ),
            )
            self._bridge.connect()
        except Exception as e:
            self._bridge = None
            raise e
        # Once bridge is connected, instantiate container using bridge id
        container_props = self._prepare_container_props(
            self._supported_intents, None
        )
        if self._default_service_timeout:
            container_props[
                ECF_PY4J_DEFAULT_SERVICE_TIMEOUT
            ] = self._default_service_timeout
        self._container = self._ipopo.instantiate(
            self._config_name, self._bridge.get_id(), container_props
        )

    @Invalidate
    def _invalidate(self, _):
        if self._bridge:
            with self._lock:
                # Set done flag to True
                self._done = True
                # Trigger reading from queue in self._worker
                # with empty task
                self._queue.put((None, None, None))
            try:
                self._ipopo.invalidate(self._bridge.get_id())
            except ValueError:
                pass
            try:
                self._bridge.disconnect()
            except Exception:
                pass
            self._bridge = None
            self._container = None

    # Implementation of Py4jServiceBridgeEventListener
    def service_imported(
        self, servicebridge, endpointid, proxy, endpoint_props
    ):
        # put on task queue so no blocking, but fifo delivery to rsa
        #  _logger.info('service_imported endpointid='+endpointid)
        self._queue.put((endpointid, endpoint_props, self._handle_import))

    def service_modified(
        self, servicebridge, endpointid, proxy, endpoint_props
    ):
        # _logger.info('_service_modified endpointid='+endpointid+";proxy="+str(proxy)+";endpoint_props="+str(endpoint_props))
        self._queue.put(
            (endpointid, endpoint_props, self._handle_import_update)
        )

    def service_unimported(
        self, servicebridge, endpointid, proxy, endpoint_props
    ):
        # _logger.info('_service_unimported endpointid='+endpointid+";proxy="+str(proxy)+";endpoint_props="+str(endpoint_props))
        # put on task queue so no blocking, but fifo delivery to rsa
        self._queue.put(
            (endpointid,
             endpoint_props,
             self._handle_import_close))

    @PostRegistration
    def _post_reg(self, _):
        # start the thread for processing import_service import requests
        self._thread.start()

    # this is method called by self._thread.  All it does is
    # read from queue, and import/unregister imported the discovered service
    def _worker(self):
        while True:
            with self._lock:
                # If self._done flag is set, return and that's it
                if self._done:
                    return
            # otherwise block to get items from queue placed by service_imported,
            # service_modified, and service_unimported
            # called by Py4j handler thread
            item = self._queue.get()
            f = None
            try:
                # get the function from item[2]
                f = item[2]
            except Exception:
                logging.error("Exception getting code in item=%s", item)

            if f:
                try:
                    # get the endpoint description properties from item[1]
                    # and create EndpointDescription instance
                    ed = EndpointDescription(properties=item[1])
                except Exception:
                    logging.error(
                        "Exception creating endpoint description from props=%s",
                        item[1],
                    )
                else:
                    # call appropriate function
                    try:
                        f(ed)
                    except Exception:
                        logging.error("Exception invoking function=%s", f)

            # no matter what, we are done with this task
            self._queue.task_done()



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Temporal dependency handler

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import threading

# Pelix beans
from pelix.constants import BundleActivator
import pelix.utilities as utilities

# iPOPO constants
import pelix.ipopo.constants as ipopo_constants
import pelix.ipopo.handlers.constants as constants
import pelix.ipopo.handlers.requires as requires

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


class _HandlerFactory(constants.HandlerFactory):
    # pylint: disable=R0903
    """
    Factory service for service registration handlers
    """

    @staticmethod
    def _prepare_configs(configs, requires_filters, temporal_timeouts):
        """
        Overrides the filters specified in the decorator with the given ones

        :param configs: Field → (Requirement, key, allow_none) dictionary
        :param requires_filters: Content of the 'requires.filter' component
                                 property (field → string)
        :param temporal_timeouts: Content of the 'temporal.timeouts' component
                                  property (field → float)
        :return: The new configuration dictionary
        """
        if not isinstance(requires_filters, dict):
            requires_filters = {}

        if not isinstance(temporal_timeouts, dict):
            temporal_timeouts = {}

        if not requires_filters and not temporal_timeouts:
            # No explicit configuration given
            return configs

        # We need to change a part of the requirements
        new_configs = {}
        for field, config in configs.items():
            # Extract values from tuple
            requirement, timeout = config
            explicit_filter = requires_filters.get(field)
            explicit_timeout = temporal_timeouts.get(field)

            # Convert the timeout value
            try:
                explicit_timeout = int(explicit_timeout)
                if explicit_timeout <= 0:
                    explicit_timeout = timeout
            except (ValueError, TypeError):
                explicit_timeout = timeout

            if not explicit_filter and not explicit_timeout:
                # Nothing to do
                new_configs[field] = config
            else:
                try:
                    # Store an updated copy of the requirement
                    requirement_copy = requirement.copy()
                    if explicit_filter:
                        requirement_copy.set_filter(explicit_filter)
                    new_configs[field] = (requirement_copy, explicit_timeout)
                except (TypeError, ValueError):
                    # No information for this one, or invalid filter:
                    # keep the factory requirement
                    new_configs[field] = config

        return new_configs

    def get_handlers(self, component_context, instance):
        """
        Sets up service providers for the given component

        :param component_context: The ComponentContext bean
        :param instance: The component instance
        :return: The list of handlers associated to the given component
        """
        # Extract information from the context
        configs = component_context.get_handler(
            ipopo_constants.HANDLER_TEMPORAL
        )
        requires_filters = component_context.properties.get(
            ipopo_constants.IPOPO_REQUIRES_FILTERS, None
        )
        temporal_timeouts = component_context.properties.get(
            ipopo_constants.IPOPO_TEMPORAL_TIMEOUTS, None
        )

        # Prepare requirements
        new_configs = self._prepare_configs(
            configs, requires_filters, temporal_timeouts
        )

        # Return handlers
        return [
            TemporalDependency(field, requirement, timeout)
            for field, (requirement, timeout) in new_configs.items()
        ]


@BundleActivator
class _Activator(object):
    """
    The bundle activator
    """

    def __init__(self):
        """
        Sets up members
        """
        self._registration = None

    def start(self, context):
        """
        Bundle started
        """
        # Set up properties
        properties = {
            constants.PROP_HANDLER_ID: ipopo_constants.HANDLER_TEMPORAL
        }

        # Register the handler factory service
        self._registration = context.register_service(
            constants.SERVICE_IPOPO_HANDLER_FACTORY,
            _HandlerFactory(),
            properties,
        )

    def stop(self, _):
        """
        Bundle stopped
        """
        # Unregister the service
        self._registration.unregister()
        self._registration = None


# ------------------------------------------------------------------------------


class TemporalException(constants.HandlerException):
    """
    Temporal exception
    """

    pass


class _TemporalProxy(object):
    """
    The injected proxy
    """

    def __init__(self, timeout):
        """
        The temporal proxy
        """
        self.__event = utilities.EventData()
        self.__timeout = timeout

    def set_service(self, service):
        """
        Sets the injected service

        :param service: The injected service, or None
        """
        self.__event.set(service)

    def unset_service(self):
        """
        The injected service has gone away
        """
        self.__event.clear()

    def __getattr__(self, item):
        """
        Returns the attribute from the "real" service

        :return: The attribute
        """
        if self.__event.wait(self.__timeout):
            return getattr(self.__event.data, item)
        else:
            raise TemporalException("No service found before timeout")

    def __call__(self, *args, **kwargs):
        """
        Call the underlying object. Lets exception propagate
        """
        if self.__event.wait(self.__timeout):
            # We have a service: call it
            return self.__event.data.__call__(*args, **kwargs)
        else:
            raise TemporalException("No service found before timeout")

    def __bool__(self):
        """
        Boolean value of the proxy
        """
        return self.__event.is_set() and bool(self.__event.data)

    # Python 2 compatibility
    __nonzero__ = __bool__


class TemporalDependency(requires.SimpleDependency):
    """
    Manages a temporal dependency field
    """

    def __init__(self, field, requirement, timeout):
        """
        Sets up the dependency

        :param field: Field where to inject the proxy
        :param requirement: Description of the required dependency
        :param timeout: Time to wait for a service (greater than 0, in seconds)
        """
        super(TemporalDependency, self).__init__(field, requirement)

        # Internal timeout
        self.__timeout = timeout

        # The delayed unbind timer
        self.__timer = None
        self.__timer_args = None
        self.__still_valid = False

        # The injected value is the proxy
        self._value = _TemporalProxy(self.__timeout)

    def clear(self):
        """
        Cleans up the manager. The manager can't be used after this method has
        been called
        """
        # Cancel timer
        self.__cancel_timer()
        self.__timer = None
        self.__timer_args = None

        self.__still_valid = False
        self._value = None
        super(TemporalDependency, self).clear()

    def on_service_arrival(self, svc_ref):
        """
        Called when a service has been registered in the framework

        :param svc_ref: A service reference
        """
        with self._lock:
            if self.reference is None:
                # Inject the service
                service = self._context.get_service(svc_ref)
                self.reference = svc_ref
                self._value.set_service(service)
                self.__still_valid = True

                # Cancel timer
                self.__cancel_timer()

                # Bind the service
                self._ipopo_instance.bind(self, self._value, self.reference)
                return True

            return None

    def on_service_departure(self, svc_ref):
        """
        Called when a service has been unregistered from the framework

        :param svc_ref: A service reference
        """
        with self._lock:
            if svc_ref is self.reference:
                # Forget about the service
                self._value.unset_service()

                # Clear the reference
                self.reference = None

                # Look for a replacement
                self._pending_ref = self._context.get_service_reference(
                    self.requirement.specification, self.requirement.filter
                )

                if self._pending_ref is None:
                    # No replacement found yet, wait a little
                    self.__still_valid = True
                    self.__timer_args = (self._value, svc_ref)
                    self.__timer = threading.Timer(
                        self.__timeout, self.__unbind_call, (False,)
                    )
                    self.__timer.start()

                else:
                    # Notify iPOPO immediately
                    self._ipopo_instance.unbind(self, self._value, svc_ref)
                return True

            return None

    def __cancel_timer(self):
        """
        Cancels the timer, and calls its target method immediately
        """
        if self.__timer is not None:
            self.__timer.cancel()
            self.__unbind_call(True)

        self.__timer_args = None
        self.__timer = None

    def __unbind_call(self, still_valid):
        """
        Calls the iPOPO unbind method
        """
        with self._lock:
            if self.__timer is not None:
                # Timeout expired, we're not valid anymore
                self.__timer = None
                self.__still_valid = still_valid
                self._ipopo_instance.unbind(
                    self, self.__timer_args[0], self.__timer_args[1]
                )

    def is_valid(self):
        """
        Tests if the dependency is in a valid state
        """
        # Don't use the parent method: it will return true as the "_value"
        # member is not None
        return (
            self.__still_valid
            or self._pending_ref is not None
            or self.requirement.optional
        )



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Dependency-less LDAP filter parser for Python

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Standard library
import inspect

# Standard typing module should be optional
try:
    # pylint: disable=W0611
    from typing import Any, Callable, Iterable, Optional, Union
except ImportError:
    pass

# Pelix
from pelix.utilities import is_string

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------

ESCAPE_CHARACTER = "\\"
""" The LDAP escape character: \\"""

# Do not insert the ESCAPE CHARACTER in this list
ESCAPED_CHARACTERS = "()&|=<>~*+#,;'\""
"""
The characters to escape in an LDAP string.
See http://www.ldapexplorer.com/en/manual/109010000-ldap-filter-syntax.htm
"""

# ------------------------------------------------------------------------------

AND = 0
""" 'And' LDAP operation """

OR = 1
""" 'Or' LDAP operation """

NOT = 2
""" 'Not' LDAP operation """

# ------------------------------------------------------------------------------


class LDAPFilter(object):
    """
    Represents an LDAP filter
    """

    __slots__ = ("subfilters", "operator")

    def __init__(self, operator):
        """
        Initializer
        """
        if operator not in (AND, OR, NOT):
            raise ValueError("Invalid operator: {0}".format(operator))

        self.subfilters = []
        self.operator = operator

    def __eq__(self, other):
        """
        Equality testing
        """
        if not isinstance(other, LDAPFilter):
            # Bad type
            return False

        if self.operator != other.operator:
            # Different operators
            return False

        # All sub-filters must match
        if len(self.subfilters) != len(other.subfilters):
            # Not the same size...
            return False

        for subfilter in self.subfilters:
            if subfilter not in other.subfilters:
                # Missing sub filter
                return False

        # Same content
        return True

    def __ne__(self, other):
        """
        Inequality testing
        """
        return not self.__eq__(other)

    def __repr__(self):
        """
        String description
        """
        return "{0}.get_ldap_filter({1!r})".format(__name__, self.__str__())

    def __str__(self):
        """
        String representation
        """
        return "({0}{1})".format(
            operator2str(self.operator),
            "".join(str(subfilter) for subfilter in self.subfilters),
        )

    def append(self, ldap_filter):
        """
        Appends a filter or a criterion to this filter

        :param ldap_filter: An LDAP filter or criterion
        :raise TypeError: If the parameter is not of a known type
        :raise ValueError: If the more than one filter is associated to a
                           NOT operator
        """
        if not isinstance(ldap_filter, (LDAPFilter, LDAPCriteria)):
            raise TypeError(
                "Invalid filter type: {0}".format(type(ldap_filter).__name__)
            )

        if len(self.subfilters) >= 1 and self.operator == NOT:
            raise ValueError("Not operator only handles one child")

        self.subfilters.append(ldap_filter)

    def matches(self, properties):
        """
        Tests if the given properties matches this LDAP filter and its children

        :param properties: A dictionary of properties
        :return: True if the properties matches this filter, else False
        """
        # Use a generator, and declare it outside of the method call
        # => seems to be quite a speed up trick
        generator = (
            criterion.matches(properties) for criterion in self.subfilters
        )

        # Extract "if" from loops and use built-in methods
        if self.operator == OR:
            result = any(generator)
        else:
            result = all(generator)
            if self.operator == NOT:
                # Revert result
                return not result

        return result

    def normalize(self):
        """
        Returns the first meaningful object in this filter.
        """
        if not self.subfilters:
            # No sub-filters
            return None

        # New sub-filters list
        new_filters = []
        for subfilter in self.subfilters:
            # Normalize the sub-filter before storing it
            norm_filter = subfilter.normalize()
            if norm_filter is not None and norm_filter not in new_filters:
                new_filters.append(norm_filter)

        # Update the instance
        self.subfilters = new_filters

        size = len(self.subfilters)
        if size > 1 or self.operator == NOT:
            # Normal filter or NOT
            # NOT is the only operator to accept 1 operand
            return self

        # Return the only child as the filter object
        return self.subfilters[0].normalize()


class LDAPCriteria(object):
    """
    Represents an LDAP criterion
    """

    __slots__ = ("name", "value", "comparator")

    def __init__(self, name, value, comparator):
        """
        Sets up the criterion

        :raise ValueError: If one of the parameters is empty
        """
        if not name or not value or not comparator:
            # Refuse empty values
            raise ValueError(
                "Invalid criterion parameter ({0}, {1}, {2})".format(
                    name, value, comparator
                )
            )

        if not (inspect.ismethod(comparator) or inspect.isfunction(comparator)):
            # Ensure we have a valid comparator
            raise ValueError(
                "Comparator must be a method: {0}".format(comparator)
            )

        self.name = str(name)
        self.value = value
        self.comparator = comparator

    def __eq__(self, other):
        """
        Equality testing
        """
        if not isinstance(other, LDAPCriteria):
            # Bad type
            return False

        for member in "name", "comparator":
            if getattr(self, member) != getattr(other, member):
                # Difference found
                return False

        if isinstance(other.value, type(self.value)):
            # Same type: direct comparison
            return self.value == other.value

        # Convert to strings for comparison
        return str(self.value) == str(other.value)

    def __ne__(self, other):
        """
        Inequality testing
        """
        return not self.__eq__(other)

    def __repr__(self):
        """
        String representation
        """
        return "{0}.get_ldap_filter({1!r})".format(__name__, self.__str__())

    def __str__(self):
        """
        String description
        """
        return "({0}{1}{2})".format(
            escape_LDAP(self.name),
            comparator2str(self.comparator),
            escape_LDAP(str(self.value)),
        )

    def matches(self, properties):
        """
        Tests if the given criterion matches this LDAP criterion

        :param properties: A dictionary of properties
        :return: True if the properties matches this criterion, else False
        """
        try:
            # Use the comparator
            return self.comparator(self.value, properties[self.name])
        except KeyError:
            # Criterion key is not in the properties
            return False

    def normalize(self):
        """
        Returns this criterion
        """
        return self


# ------------------------------------------------------------------------------


def escape_LDAP(ldap_string):
    # type: (str) -> str
    # pylint: disable=C0103
    """
    Escape a string to let it go in an LDAP filter

    :param ldap_string: The string to escape
    :return: The protected string
    """
    if not ldap_string:
        # No content
        return ldap_string

    # Protect escape character previously in the string
    assert is_string(ldap_string)
    ldap_string = ldap_string.replace(
        ESCAPE_CHARACTER, ESCAPE_CHARACTER + ESCAPE_CHARACTER
    )

    # Leading space
    if ldap_string.startswith(" "):
        ldap_string = "\\ {0}".format(ldap_string[1:])

    # Trailing space
    if ldap_string.endswith(" "):
        ldap_string = "{0}\\ ".format(ldap_string[:-1])

    # Escape other characters
    for escaped in ESCAPED_CHARACTERS:
        ldap_string = ldap_string.replace(escaped, ESCAPE_CHARACTER + escaped)

    return ldap_string


def unescape_LDAP(ldap_string):
    # type: (str) -> str
    # pylint: disable=C0103
    """
    Unespaces an LDAP string

    :param ldap_string: The string to unescape
    :return: The unprotected string
    """
    if ldap_string is None:
        return None

    if ESCAPE_CHARACTER not in ldap_string:
        # No need to loop
        return ldap_string

    escaped = False
    result = ""

    for character in ldap_string:
        if not escaped and character == ESCAPE_CHARACTER:
            # Escape character found
            escaped = True
        else:
            # Copy the character
            escaped = False
            result += character

    return result


# ------------------------------------------------------------------------------


ITERABLES = (list, tuple, set, frozenset)
""" The types that are considered iterable in comparators """


def _comparator_presence(_, tested_value):
    """
    Tests a filter which simply a joker, i.e. a value presence test
    """
    # The filter value is a joker : simple presence test
    if tested_value is None:
        return False
    elif hasattr(tested_value, "__len__"):
        # Refuse empty values
        # pylint: disable=C1801
        return len(tested_value) != 0

    # Presence validated
    return True


def _comparator_star(filter_value, tested_value):
    """
    Tests a filter containing a joker
    """
    if isinstance(tested_value, ITERABLES):
        for value in tested_value:
            if _star_comparison(filter_value, value):
                return True
        return False

    return _star_comparison(filter_value, tested_value)


def _star_comparison(filter_value, tested_value):
    """
    Tests a filter containing a joker
    """
    if not is_string(tested_value):
        # Unhandled value type...
        return False

    parts = filter_value.split("*")

    i = 0
    last_part = len(parts) - 1

    idx = 0
    for part in parts:
        # Find the part in the tested value
        idx = tested_value.find(part, idx)
        if idx == -1:
            # Part not found
            return False

        len_part = len(part)
        if i == 0 and len_part != 0 and idx != 0:
            # First part is not a star, but the tested value is not at
            # position 0 => Doesn't match
            return False

        if (
            i == last_part
            and len_part != 0
            and idx != len(tested_value) - len_part
        ):
            # Last tested part is not at the end of the sequence
            return False

        # Be sure to test the next part
        idx += len_part
        i += 1

    # Whole test passed
    return True


def _comparator_eq(filter_value, tested_value):
    """
    Tests if the filter value is equal to the tested value
    """
    if isinstance(tested_value, ITERABLES):
        # Convert the list items to strings
        for value in tested_value:
            # Try with the string conversion
            if not is_string(value):
                value = repr(value)

            if filter_value == value:
                # Match !
                return True
    # Standard comparison
    elif not is_string(tested_value):
        # String vs string representation
        return filter_value == repr(tested_value)
    else:
        # String vs string
        return filter_value == tested_value

    return False


def _comparator_approximate(filter_value, tested_value):
    """
    Tests if the filter value is nearly equal to the tested value.

    If the tested value is a string or an array of string, it compares their
    lower case forms
    """
    lower_filter_value = filter_value.lower()

    if is_string(tested_value):
        # Lower case comparison
        return _comparator_eq(lower_filter_value, tested_value.lower())

    elif hasattr(tested_value, "__iter__"):
        # Extract a list of strings
        new_tested = [
            value.lower() for value in tested_value if is_string(value)
        ]

        if _comparator_eq(lower_filter_value, new_tested):
            # Value found in the strings
            return True

    # Compare the raw values
    return _comparator_eq(filter_value, tested_value) or _comparator_eq(
        lower_filter_value, tested_value
    )


def _comparator_approximate_star(filter_value, tested_value):
    """
    Tests if the filter value, which contains a joker, is nearly equal to the
    tested value.

    If the tested value is a string or an array of string, it compares their
    lower case forms
    """
    lower_filter_value = filter_value.lower()

    if is_string(tested_value):
        # Lower case comparison
        return _comparator_star(lower_filter_value, tested_value.lower())

    elif hasattr(tested_value, "__iter__"):
        # Extract a list of strings
        new_tested = [
            value.lower() for value in tested_value if is_string(value)
        ]

        if _comparator_star(lower_filter_value, new_tested):
            # Value found in the strings
            return True

    # Compare the raw values
    return _comparator_star(filter_value, tested_value) or _comparator_star(
        lower_filter_value, tested_value
    )


def _comparator_le(filter_value, tested_value):
    """
    Tests if the filter value is greater than the tested value

    tested_value <= filter_value
    """
    return _comparator_lt(filter_value, tested_value) or _comparator_eq(
        filter_value, tested_value
    )


def _comparator_lt(filter_value, tested_value):
    """
    Tests if the filter value is strictly greater than the tested value

    tested_value < filter_value
    """
    if is_string(filter_value):
        value_type = type(tested_value)
        try:
            # Try a conversion
            filter_value = value_type(filter_value)

        except (TypeError, ValueError):
            if value_type is int:
                # Integer/float comparison trick
                try:
                    filter_value = float(filter_value)
                except (TypeError, ValueError):
                    # None-float value
                    return False
            else:
                # Incompatible type
                return False
    try:
        return tested_value < filter_value
    except TypeError:
        # Incompatible type
        return False


def _comparator_ge(filter_value, tested_value):
    """
    Tests if the filter value is lesser than the tested value

    tested_value >= filter_value
    """
    return _comparator_gt(filter_value, tested_value) or _comparator_eq(
        filter_value, tested_value
    )


def _comparator_gt(filter_value, tested_value):
    """
    Tests if the filter value is strictly lesser than the tested value

    tested_value > filter_value
    """
    if is_string(filter_value):
        value_type = type(tested_value)
        try:
            # Try a conversion
            filter_value = value_type(filter_value)
        except (TypeError, ValueError):
            if value_type is int:
                # Integer/float comparison trick
                try:
                    filter_value = float(filter_value)
                except (TypeError, ValueError):
                    # None-float value
                    return False
            else:
                # Incompatible type
                return False
    try:
        return tested_value > filter_value
    except TypeError:
        # Incompatible type
        return False


_COMPARATOR_SYMBOL = {
    _comparator_approximate: "~=",
    _comparator_approximate_star: "~=",
    _comparator_eq: "=",
    _comparator_star: "=",
    _comparator_le: "<=",
    _comparator_lt: "<",
    _comparator_ge: ">=",
    _comparator_gt: ">",
}


def comparator2str(comparator):
    # type: (Callable[[Any, Any], bool]) -> str
    """
    Converts an operator method to a string

    :param comparator: A comparator method
    :return: The corresponding LDAP filter comparator string
    """
    return _COMPARATOR_SYMBOL.get(comparator, "??")


def operator2str(operator):
    # type: (int) -> str
    """
    Converts an operator value to a string

    :param operator: An LDAP filter operator internal value
    :return: The corresponding LDAP operator string
    """
    if operator == AND:
        return "&"
    elif operator == OR:
        return "|"
    elif operator == NOT:
        return "!"
    return "<unknown>"


# ------------------------------------------------------------------------------


def _compute_comparator(string, idx):
    # type: (str, int) -> Optional[Callable[[Any, Any], bool]]
    """
    Tries to compute the LDAP comparator at the given index

    Valid operators are :

    * = : equality
    * <= : less than
    * >= : greater than
    * ~= : approximate

    :param string: A LDAP filter string
    :param idx: An index in the given string
    :return: The corresponding operator, None if unknown
    """
    part1 = string[idx]
    try:
        part2 = string[idx + 1]
    except IndexError:
        # String is too short (no comparison)
        return None

    if part1 == "=":
        # Equality
        return _comparator_eq
    elif part2 != "=":
        # It's a "strict" operator
        if part1 == "<":
            # Strictly lesser
            return _comparator_lt
        elif part1 == ">":
            # Strictly greater
            return _comparator_gt
    else:
        if part1 == "<":
            # Less or equal
            return _comparator_le
        elif part1 == ">":
            # Greater or equal
            return _comparator_ge
        elif part1 == "~":
            # Approximate equality
            return _comparator_approximate
    return None


def _compute_operation(string, idx):
    # type: (str, int) -> Optional[int]
    """
    Tries to compute the LDAP operation at the given index

    Valid operations are :

    * & : AND
    * | : OR
    * ! : NOT

    :param string: A LDAP filter string
    :param idx: An index in the given string
    :return: The corresponding operator (AND, OR or NOT)
    """
    operator = string[idx]
    if operator == "&":
        return AND
    elif operator == "|":
        return OR
    elif operator == "!":
        return NOT

    return None


def _skip_spaces(string, idx):
    # type: (str, int) -> int
    """
    Retrieves the next non-space character after idx index in the given string

    :param string: The string to look into
    :param idx: The base search index
    :return: The next non-space character index, -1 if not found
    """
    i = idx
    for char in string[idx:]:
        if not char.isspace():
            return i
        i += 1

    return -1


def _parse_ldap_criteria(ldap_filter, startidx=0, endidx=-1):
    # type: (str, int, int) -> LDAPCriteria
    """
    Parses an LDAP sub filter (criterion)

    :param ldap_filter: An LDAP filter string
    :param startidx: Sub-filter start index
    :param endidx: Sub-filter end index
    :return: The LDAP sub-filter
    :raise ValueError: Invalid sub-filter
    """
    comparators = "=<>~"
    if startidx < 0:
        raise ValueError(
            "Invalid string range start={0}, end={1}".format(startidx, endidx)
        )

    # Get the comparator
    escaped = False
    idx = startidx
    for char in ldap_filter[startidx:endidx]:
        if not escaped:
            if char == ESCAPE_CHARACTER:
                # Next character escaped
                escaped = True
            elif char in comparators:
                # Comparator found
                break
        else:
            # Escaped character ignored
            escaped = False
        idx += 1
    else:
        # Comparator never found
        raise ValueError(
            "Comparator not found in '{0}'".format(ldap_filter[startidx:endidx])
        )

    # The attribute name can be extracted directly
    attribute_name = ldap_filter[startidx:idx].strip()
    if not attribute_name:
        # Attribute name is missing
        raise ValueError(
            "Attribute name is missing in '{0}'".format(
                ldap_filter[startidx:endidx]
            )
        )

    comparator = _compute_comparator(ldap_filter, idx)
    if comparator is None:
        # Unknown comparator
        raise ValueError(
            "Unknown comparator in '{0}' - {1}\nFilter : {2}".format(
                ldap_filter[startidx:endidx], ldap_filter[idx], ldap_filter
            )
        )

    # Find the end of the comparator
    while ldap_filter[idx] in comparators:
        idx += 1

    # Skip spaces
    idx = _skip_spaces(ldap_filter, idx)

    # Extract the value
    value = ldap_filter[idx:endidx].strip()

    # Use the appropriate comparator if a joker is found in the filter value
    if value == "*":
        # Presence comparator
        comparator = _comparator_presence
    elif "*" in value:
        # Joker
        if comparator == _comparator_eq:
            comparator = _comparator_star
        elif comparator == _comparator_approximate:
            comparator = _comparator_approximate_star

    return LDAPCriteria(
        unescape_LDAP(attribute_name), unescape_LDAP(value), comparator
    )


def _parse_ldap(ldap_filter):
    # type: (str) -> Optional[LDAPFilter]
    """
    Parses the given LDAP filter string

    :param ldap_filter: An LDAP filter string
    :return: An LDAPFilter object, None if the filter was empty
    :raise ValueError: The LDAP filter string is invalid
    """
    if ldap_filter is None:
        # Nothing to do
        return None

    assert is_string(ldap_filter)

    # Remove surrounding spaces
    ldap_filter = ldap_filter.strip()
    if not ldap_filter:
        # Empty string
        return None

    escaped = False
    filter_len = len(ldap_filter)
    root = None
    stack = []
    subfilter_stack = []

    idx = 0
    while idx < filter_len:
        if not escaped:
            if ldap_filter[idx] == "(":
                # Opening filter : get the operator
                idx = _skip_spaces(ldap_filter, idx + 1)
                if idx == -1:
                    raise ValueError(
                        "Missing filter operator: {0}".format(ldap_filter)
                    )

                operator = _compute_operation(ldap_filter, idx)
                if operator is not None:
                    # New sub-filter
                    stack.append(LDAPFilter(operator))
                else:
                    # Sub-filter content
                    subfilter_stack.append(idx)

            elif ldap_filter[idx] == ")":
                # Ending filter : store it in its parent
                if subfilter_stack:
                    # criterion finished
                    start_idx = subfilter_stack.pop()
                    criterion = _parse_ldap_criteria(
                        ldap_filter, start_idx, idx
                    )

                    if stack:
                        top = stack.pop()
                        top.append(criterion)
                        stack.append(top)
                    else:
                        # No parent : filter contains only one criterion
                        # Make a parent to stay homogeneous
                        root = LDAPFilter(AND)
                        root.append(criterion)
                elif stack:
                    # Sub filter finished
                    ended_filter = stack.pop()
                    if stack:
                        top = stack.pop()
                        top.append(ended_filter)
                        stack.append(top)
                    else:
                        # End of the parse
                        root = ended_filter
                else:
                    raise ValueError(
                        "Too many end of parenthesis:{0}: {1}".format(
                            idx, ldap_filter[idx:]
                        )
                    )
            elif ldap_filter[idx] == "\\":
                # Next character must be ignored
                escaped = True
        else:
            # Escaped character ignored
            escaped = False

        # Don't forget to increment...
        idx += 1

    # No root : invalid content
    if root is None:
        raise ValueError("Invalid filter string: {0}".format(ldap_filter))

    # Return the root of the filter
    return root.normalize()


def get_ldap_filter(ldap_filter):
    # type: (Any) -> Optional[Union[LDAPFilter, LDAPCriteria]]
    """
    Retrieves the LDAP filter object corresponding to the given filter.
    Parses it the argument if it is an LDAPFilter instance

    :param ldap_filter: An LDAP filter (LDAPFilter or string)
    :return: The corresponding filter, can be None
    :raise ValueError: Invalid filter string found
    :raise TypeError: Unknown filter type
    """
    if ldap_filter is None:
        return None

    if isinstance(ldap_filter, (LDAPFilter, LDAPCriteria)):
        # No conversion needed
        return ldap_filter
    elif is_string(ldap_filter):
        # Parse the filter
        return _parse_ldap(ldap_filter)

    # Unknown type
    raise TypeError(
        "Unhandled filter type {0}".format(type(ldap_filter).__name__)
    )


def combine_filters(filters, operator=AND):
    # type: (Iterable[Any], int) -> Optional[Union[LDAPFilter, LDAPCriteria]]
    """
    Combines two LDAP filters, which can be strings or LDAPFilter objects

    :param filters: Filters to combine
    :param operator: The operator for combination
    :return: The combined filter, can be None if all filters are None
    :raise ValueError: Invalid filter string found
    :raise TypeError: Unknown filter type
    """
    if not filters:
        return None

    if not hasattr(filters, "__iter__") or is_string(filters):
        raise TypeError("Filters argument must be iterable")

    # Remove None filters and convert others
    ldap_filters = []
    for sub_filter in filters:
        if sub_filter is None:
            # Ignore None filters
            continue

        ldap_filter = get_ldap_filter(sub_filter)
        if ldap_filter is not None:
            # Valid filter
            ldap_filters.append(ldap_filter)

    if not ldap_filters:
        # Do nothing
        return None
    elif len(ldap_filters) == 1:
        # Only one filter, return it
        return ldap_filters[0]

    new_filter = LDAPFilter(operator)
    for sub_filter in ldap_filters:
        # Direct combination
        new_filter.append(sub_filter)

    return new_filter.normalize()



#!/usr/bin/env python
# -- Content-Encoding: UTF-8 --
"""
Configuration Admin shell commands

Provides commands to the Pelix shell to work with the Configuration Admin
service

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Shell constants
from pelix.shell import SERVICE_SHELL_COMMAND

# iPOPO Decorators
from pelix.ipopo.decorators import (
    ComponentFactory,
    Requires,
    Provides,
    Instantiate,
    Invalidate,
)
import pelix.services

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# -----------------------------------------------------------------------------


@ComponentFactory("configadmin-shell-commands-factory")
@Requires("_config_admin", pelix.services.SERVICE_CONFIGURATION_ADMIN)
@Provides(SERVICE_SHELL_COMMAND)
@Instantiate("configadmin-shell-commands")
class ConfigAdminCommands(object):
    """
    Configuration Admin shell commands
    """

    def __init__(self):
        """
        Sets up members
        """
        # Injected services
        self._config_admin = None

        # Handled configurations (PID -> Configuration)
        self._configs = {}

    @Invalidate
    def invalidate(self, _):
        """
        Component invalidated
        """
        # Clean up
        self._configs.clear()

    @staticmethod
    def get_namespace():
        """
        Retrieves the name space of this command handler
        """
        return "config"

    def get_methods(self):
        """
        Retrieves the list of tuples (command, method) for this command handler
        """
        return [
            ("create", self.create),
            ("update", self.update),
            ("reload", self.reload),
            ("delete", self.delete),
            ("list", self.list),
        ]

    def create(self, io_handler, factory_pid, **kwargs):
        """
        Creates a factory configuration
        """
        config = self._config_admin.create_factory_configuration(factory_pid)

        # Print the configuration PID
        pid = config.get_pid()
        io_handler.write_line("New configuration: {0}", pid)

        if kwargs:
            # Update it immediately if some properties are already set
            config.update(kwargs)

    def update(self, _, pid, **kwargs):
        """
        Updates a configuration
        """
        # Get the configuration with given PID
        self._configs[pid] = config = self._config_admin.get_configuration(pid)

        # Get previous values
        old_properties = config.get_properties()
        if old_properties is None:
            new_properties = {}
        else:
            new_properties = old_properties.copy()

        # Update properties
        new_properties.update(kwargs)

        # Remove properties which value is now None
        for key, value in kwargs.items():
            if value == "None":
                del new_properties[key]

        # Update configuration
        config.update(new_properties)

    def reload(self, io_handler, pid):
        """
        Reloads the configuration with the given PID from the persistence
        """
        # Get the configuration with given PID
        self._configs[pid] = config = self._config_admin.get_configuration(pid)

        try:
            # Reload the file
            config.reload()
        except Exception as ex:
            # Log errors
            io_handler.write_line("Error reloading {0}: {1}", pid, ex)

    def delete(self, _, pid):
        """
        Deletes a configuration
        """
        self._config_admin.get_configuration(pid).delete()

        try:
            del self._configs[pid]
        except KeyError:
            # Configuration was unknown
            pass

    def list(self, io_handler, pid=None):
        """
        Lists known configurations
        """
        configs = self._config_admin.list_configurations()
        if not configs:
            io_handler.write_line("No configuration.")
            return

        # Filter with PID
        if pid is not None:
            for config in configs:
                if config.get_pid() == pid:
                    configs = [config]
                    break

            else:
                io_handler.write_line("No configuration with PID {0}.", pid)
                return

        lines = []
        for config in configs:
            lines.append("* {0}:".format(config.get_pid()))
            factory_pid = config.get_factory_pid()
            if factory_pid:
                lines.append("\tFactory PID: {0}".format(factory_pid))
            lines.append("\tLocation: {0}".format(config.get_bundle_location()))

            try:
                properties = config.get_properties()
                if properties is None:
                    lines.append("\tNot yet updated")

                else:
                    lines.append("\tProperties:")
                    lines.extend(
                        "\t\t{0} = {1}".format(key, value)
                        for key, value in properties.items()
                    )

            except ValueError:
                lines.append("\t** Deleted **")

        lines.append("")
        io_handler.write_line("{0}", "\n".join(lines))



#!/usr/bin/python
# -- Content-Encoding: UTF-8 --
"""
Defines the decorators associated shell completion handlers to a shell function

:author: Thomas Calmant
:copyright: Copyright 2023, Thomas Calmant
:license: Apache License 2.0
:version: 1.0.2
:status: Alpha

..

    Copyright 2020 Thomas Calmant

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

# Add some typing
try:
    # pylint: disable=W0611
    from typing import List, Dict
except ImportError:
    pass

try:
    # Everything here relies on readline
    import readline
except ImportError:
    readline = None

# ------------------------------------------------------------------------------

# Module version
__version_info__ = (1, 0, 2)
__version__ = ".".join(str(x) for x in __version_info__)

# Documentation strings format
__docformat__ = "restructuredtext en"

# ------------------------------------------------------------------------------


SVC_COMPLETER = "pelix.shell.completer"
""" Specification of a completer service """

PROP_COMPLETER_ID = "completer.id"
""" Completer service property: ID of the completer """

ATTR_COMPLETERS = "__pelix_shell_completers__"
""" Attribute added to methods to keep the list of completers """

# ------------------------------------------------------------------------------

DUMMY = "dummy"
""" Completer ID: a completer that does nothing """

BUNDLE = "pelix.bundle"
""" Completer ID: Pelix Bundle ID completer """

SERVICE = "pelix.service"
""" Completer ID: Pelix Service ID completer """

FACTORY = "ipopo.factory"
""" Completer ID: iPOPO Factory Name completer """

FACTORY_PROPERTY = "ipopo.factory.property"
""" Completer ID: iPOPO Property Name completer """

COMPONENT = "ipopo.component"
""" Completer ID: iPOPO Component Name completer """

# ------------------------------------------------------------------------------


class CompletionInfo:
    """
    Keep track of the configuration of a completion
    """

    __slots__ = ("__completers", "__multiple")

    def __init__(self, completers, multiple):
        # type: (List[str], bool) -> None
        """
        :param completers: The list of completers IDs
        :param multiple: Flag indicating the repetition of the last completer
        """
        self.__completers = completers or []
        self.__multiple = multiple

    @property
    def completers(self):
        """
        List of IDs of shell completers
        """
        return self.__completers[:]

    @property
    def multiple(self):
        """
        Flag indicating of the last completer can be reused multiple times
        """
        return self.__multiple


class Completion:
    # pylint: disable=R0903
    """
    Decorator that sets up the arguments completion of a shell method
    """

    def __init__(self, *completers, **kwargs):
        # type: (List[str], Dict[str, bool]) -> None
        """
        :param completers: A list of IDs (str) of argument completers
        :param multiple: If True, the last completer is reused multiple times
        """
        self._completers = completers  # type: List[str]
        self._multiple = kwargs.get("multiple", False)

    def __call__(self, method):
        """
        Adds a CoreCompleter with the list of completers as metadata of the
        method.
        Does nothing if the readline library isn't available

        :param method: Decorated method
        :return: The method with a new metadata for the Pelix Shell
        """
        if readline is not None and self._completers:
            # No need to setup completion if readline isn't available
            setattr(
                method,
                ATTR_COMPLETERS,
                CompletionInfo(self._completers, self._multiple),
            )

        return method
